#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import sys
from endesive import pdf, hsm

import os
import sys
import sysconfig
import datetime

os.environ['SOFTHSM2_CONF'] = 'softhsm2.conf'
if not os.path.exists(os.path.join(os.getcwd(), 'softhsm2.conf')):
    open('softhsm2.conf', 'wt').write('''\
log.level = DEBUG
directories.tokendir = %s/softhsm2/
objectstore.backend = file
slots.removable = false
''' % os.getcwd())
if not os.path.exists(os.path.join(os.getcwd(), 'softhsm2')):
    os.mkdir(os.path.join(os.getcwd(), 'softhsm2'))

#
#!/bin/bash
#SOFTHSM2_CONF=softhsm2.conf
#softhsm2-util --label "endesive" --slot 1 --init-token --pin secret1 --so-pin secret2
#softhsm2-util --show-slots
#

if sys.platform == 'win32':
    dllpath = r'W:\binw\SoftHSM2\lib\softhsm2-x64.dll'
else:
    dllpath = os.path.join(sysconfig.get_config_var('LIBDIR'), "softhsm/libsofthsm2.so")

import PyKCS11 as PK11

class Signer(hsm.HSM):
    def certificate(self):
        self.login("endesieve", "secret1")
        keyid = bytes((0x66,0x66,0x90))
        try:
            pk11objects = self.session.findObjects([(PK11.CKA_CLASS, PK11.CKO_CERTIFICATE)])
            all_attributes = [
                #PK11.CKA_SUBJECT,
                PK11.CKA_VALUE,
                #PK11.CKA_ISSUER,
                #PK11.CKA_CERTIFICATE_CATEGORY,
                #PK11.CKA_END_DATE,
                PK11.CKA_ID,
            ]

            for pk11object in pk11objects:
                try:
                    attributes = self.session.getAttributeValue(pk11object, all_attributes)
                except PK11.PyKCS11Error as e:
                    continue

                attrDict = dict(list(zip(all_attributes, attributes)))
                cert = bytes(attrDict[PK11.CKA_VALUE])
                if keyid == bytes(attrDict[PK11.CKA_ID]):
                    return keyid, cert
        finally:
            self.logout()
        return None, None

    def sign(self, keyid, data, mech):
        self.login("endesieve", "secret1")
        try:
            privKey = self.session.findObjects([(PK11.CKA_CLASS, PK11.CKO_PRIVATE_KEY), (PK11.CKA_ID, keyid)])[0]
            mech = getattr(PK11, 'CKM_%s_RSA_PKCS' % mech.upper())
            sig = self.session.sign(privKey, data, PK11.Mechanism(mech, None))
            return bytes(sig)
        finally:
            self.logout()

def main():
    date = datetime.datetime.utcnow() - datetime.timedelta(hours=12)
    date = date.strftime('D:%Y%m%d%H%M%S+00\'00\'')
    class User:
        full_name = 'u.full: ąćęłńóśżź'
        email = 'u.email: zażółcić gęślą jaźń'
        company = 'u.comp: ĄĆĘŁŃÓŚŻŹ'
        company_full_name = 'u.comp_full: ZAŻÓŁCIĆ GĘŚLĄ JAŹŃ'
    user = User()
    dct = {
        'aligned': 0,
        'sigflags': 3,
        'sigflagsft': 132,
        'sigpage': 0,
        'sigbutton': False,
        'sigfield': 'Signature-1667820612.078739',
        'auto_sigfield': False,
        'sigandcertify': False,
        'signaturebox': [175.79446979865773, 294.7236779911374, 447.47683221476507, 573.2810782865583],
        'contact': '',
        'location': '',
        'reason': '',
        'signingdate': "D:20221107123012+00'00'",
        'signature_appearance': {
            'background': [0.75, 0.8, 0.95],
            'outline': [0.2, 0.3, 0.5],
            'border': 1,
            'labels': True,
            'display': ['date']
        }
    }

    clshsm = Signer(dllpath)
    fname = 'pdf.pdf'
    if len (sys.argv) > 1:
        fname = sys.argv[1]
    datau = open(fname, 'rb').read()
    datas = pdf.cms.sign(datau, dct,
        None, None,
        [],
        'sha256',
        clshsm,
    )
    fname = fname.replace('.pdf', '-signed-cms-hsm-signature_appearance.pdf')
    with open(fname, 'wb') as fp:
        fp.write(datau)
        fp.write(datas)


main()



from endesive.pdf.PyPDF2 import generic as pdf


def makeObject(obj):
    if isinstance(obj, pdf.PdfObject):
        return obj
    if isinstance(obj, int):
        return pdf.NumberObject(obj)
    if isinstance(obj, float):
        return pdf.NumberObject(obj)
    if isinstance(obj, str):
        return pdf.createStringObject(obj)
    if isinstance(obj, (list, tuple)):
        result = pdf.ArrayObject()
        for v in obj:
            v = makeObject(v)
            result.append(v)
        return result
    if isinstance(obj, dict):
        result = PdfDict()
        for k, v in obj.items():
            v = makeObject(v)
            result[k] = v
        return result
    raise ValueError("can`t convert to PdfObject", obj)


class PdfDict(pdf.DictionaryObject):
    indirect = False
    stream = None

    def __init__(self, *args, **kwargs):
        super(PdfDict, self).__init__({})
        for k, v in kwargs.items():
            self[k] = v

    def __setitem__(self, k, v):
        if k == "stream":
            if isinstance(v, str):
                v = v.encode("latin1")
            self.stream = v
            return
        if not isinstance(k, pdf.NameObject):
            k = PdfName(k)
        v = makeObject(v)
        super(PdfDict, self).__setitem__(k, v)


class IndirectPdfDict(PdfDict):
    indirect = True


def PdfName(name):
    return pdf.NameObject("/" + name)


def PdfString(s):
    return pdf.createStringObject(s)


def PdfArray(l):
    result = pdf.ArrayObject([])
    for v in l:
        v = makeObject(v)
        result.append(v)
    return result



#!/usr/bin/env vpython3
from mako.template import Template
from mako.runtime import Context
from io import StringIO

raw_html = '<h1>Hello, ${name}!</h1>'
data = {'name': 'Tapan'}
pdf_name = 'pdf-wkhtml.pdf'

mytemplate = Template(raw_html)
buf = StringIO()
ctx = Context(buf, **data)
mytemplate.render_context(ctx)
html = buf.getvalue()

import pdfkit

WKHTMLTOPDF_OPTIONS = {
    'page-size': 'A4',
    'encoding': 'UTF-8',
}

pdfkit.from_string(
    html,
    pdf_name,
    options=WKHTMLTOPDF_OPTIONS
)



# -*- coding: utf-8 -*-
"""
    Font Metrics Utils
    ~~~~~~~~~~

    :copyright: Copyright 2019 Autodesk, Inc.
    :license: MIT, see LICENSE for details.
"""
import attr

from .validation import Number, List, Dict


@attr.s
class FontMetrics:
    """
    Class to hold our font metric calculations.
    """
    italicAngle = Number(default=0)
    usWeightClass = Number(default=500)
    isFixedPitch = Number(default=0)

    unitsPerEm = Number(default=1000)
    scale = Number(default=float(1))
    bbox = List(default=[])

    ascent = Number(default=None)
    descent = Number(default=None)
    capHeight = Number(default=None)

    stemV = Number(default=None)
    defaultWidth = Number(default=None)
    widths = List(default=[])
    cmap = Dict(default={})

    @property
    def flags(self):
        """
        See Section 9.8.2 - Font Descriptor Flags of PDF 1.7 Spec
        Bit 1 - FixedPitch
        Bit 2 - Serif
        Bit 3 - Symbolic
        Bit 4 - Script
        Bit 6 - Nonsymbolic
        Bit 7 - Italic
        Bit 17 - AllCap
        Bit 18 - SmallCap
        Bit 19 - ForceBold
        :return:
        """
        flags = 4
        if self.italicAngle != 0:
            flags = flags | 64
        if self.usWeightClass >= 600:
            flags = flags | 262144
        if self.isFixedPitch:
            flags = flags | 1
        return flags



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
from certum import dllpath
from lxml import etree
import PyKCS11 as PK11
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from endesive import xades, signer, hsm



class Signer(hsm.HSM):
    def certificate(self):
        self.login("profil bezpieczny", "9593")
        keyid = [
            0x5E,
            0x9A,
            0x33,
            0x44,
            0x8B,
            0xC3,
            0xA1,
            0x35,
            0x33,
            0xC7,
            0xC2,
            0x02,
            0xF6,
            0x9B,
            0xDE,
            0x55,
            0xFE,
            0x83,
            0x7B,
            0xDE,
        ]
        # keyid = [0x3f, 0xa6, 0x63, 0xdb, 0x75, 0x97, 0x5d, 0xa6, 0xb0, 0x32, 0xef, 0x2d, 0xdc, 0xc4, 0x8d, 0xe8]
        keyid = bytes(keyid)
        try:
            pk11objects = self.session.findObjects(
                [(PK11.CKA_CLASS, PK11.CKO_CERTIFICATE)]
            )
            all_attributes = [
                # PK11.CKA_SUBJECT,
                PK11.CKA_VALUE,
                # PK11.CKA_ISSUER,
                # PK11.CKA_CERTIFICATE_CATEGORY,
                # PK11.CKA_END_DATE,
                PK11.CKA_ID,
            ]

            for pk11object in pk11objects:
                try:
                    attributes = self.session.getAttributeValue(
                        pk11object, all_attributes
                    )
                except PK11.PyKCS11Error as e:
                    continue

                attrDict = dict(list(zip(all_attributes, attributes)))
                cert = bytes(attrDict[PK11.CKA_VALUE])
                if keyid == bytes(attrDict[PK11.CKA_ID]):
                    return keyid, cert
        finally:
            self.logout()
        return None, None

    def sign(self, keyid, data, mech):
        self.login("profil bezpieczny", "9593")
        try:
            privKey = self.session.findObjects(
                [(PK11.CKA_CLASS, PK11.CKO_PRIVATE_KEY), (PK11.CKA_ID, keyid)]
            )[0]
            mech = getattr(PK11, "CKM_%s_RSA_PKCS" % mech.upper())
            sig = self.session.sign(privKey, data, PK11.Mechanism(mech, None))
            return bytes(sig)
        finally:
            self.logout()


def main():
    clshsm = Signer(dllpath)
    keyid, cert = clshsm.certificate()

    def signproc(tosign, algosig):
        return clshsm.sign(keyid, tosign, algosig)

    data = open("xml.xml", "rb").read()
    cert = x509.load_der_x509_certificate(cert, backend=default_backend())
    certcontent = cert.public_bytes(serialization.Encoding.DER)

    for tspurl, tspcred in (
        (None, None),
        ("http://public-qlts.certum.pl/qts-17", None)
    ):
        cls = xades.BES()
        doc = cls.enveloping(
            "dokument.xml",
            data,
            "application/xml",
            cert,
            certcontent,
            signproc,
            False,
            False,
            False,
            tspurl,
            tspcred,
        )
        data = etree.tostring(doc, encoding="UTF-8", xml_declaration=True, standalone=False)
        if tspurl is None:
            open("xml-hsm-certum-enveloping.xml", "wb").write(data)
        else:
            open("xml-hsm-certum-enveloping-t.xml", "wb").write(data)


if __name__ == "__main__":
    main()



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import sys
import datetime
from cryptography import x509
from cryptography.hazmat import backends
from cryptography.hazmat.primitives.serialization import pkcs12
from endesive import pdf

# import logging
# logging.basicConfig(level=logging.DEBUG)


def main():
    date = datetime.datetime.utcnow() - datetime.timedelta()
    date = date.strftime("%Y%m%d%H%M%S+00'00'")
    dct = {
        "sigflags": 3,
        "sigpage": 0,
        "sigbutton": True,
        "contact": "mak@trisoft.com.pl",
        "location": "Szczecin",
        "signingdate": date.encode(),
        "reason": "Dokument podpisany cyfrowo",
        "signature": "Dokument podpisany cyfrowo",
        "signaturebox": (0, 0, 100, 100),
        "sigandcertify": True,
        "text": {
            "fontsize": 10,
        },
    }

    pk12fname = "/home/mak/Dokumenty/m32/ssl/actalis/actalis.p12"
    pk12pass = sys.argv[1].encode()
    with open(pk12fname, "rb") as fp:
        p12 = pkcs12.load_key_and_certificates(
            fp.read(), pk12pass, backends.default_backend()
        )

    fname = "pdf.pdf"
    if len(sys.argv) > 2:
        fname = sys.argv[2]
    datau = open(fname, "rb").read()
    datas = pdf.cms.sign(
        datau,
        dct,
        p12[0],
        p12[1],
        p12[2][:3],
        "sha256",
        None,
    )
    fname = fname.replace(".pdf", "-signed-cms-m32-actalis.pdf")
    with open(fname, "wb") as fp:
        fp.write(datau)
        fp.write(datas)


main()



# *-* coding: utf-8 *-*
import hashlib

from endesive import signer
from . import fpdf


class FPDF(fpdf.FPDF):
    signer = True

    def pkcs11_aligned(self, data):
        data = ''.join(['%02x' % i for i in data])
        nb = 0x4000 - len(data)
        data = data + '0' * (0x4000 - len(data))
        return data

    def pkcs11_setup(self, config, key, cert, othercerts, algomd):
        self.pkcs11config = config
        self.pkcs11zeros = self.pkcs11_aligned([0])
        self.pkcs11annot = 0
        self.pkcs11key = key
        self.pkcs11cert = cert
        self.pkcs11certs = othercerts
        self.pkcs11algomd = algomd

    def pkcs11_signature(self):
        self._newobj()
        self.pkcs11annot = self.n
        self._out('<</F 132/Type/Annot/Subtype/Widget/Rect[0 0 0 0]/FT/Sig/DR<<>>/T(signature%d)/V %d 0 R>>' % (
        self.pkcs11annot, self.pkcs11annot + 1))
        self._out('endobj')

        self._newobj()
        self._out('<</Type/Sig/SubFilter/adbe.pkcs7.detached/Location(%(location)s)/M(D:%(signingdate)s)' % self.pkcs11config)
        self._out(
            '/ByteRange [0000000000 0000000000 0000000000 0000000000]/Filter/Adobe.PPKLite/Reason(%(reason)s)/ContactInfo(%(contact)s)' % self.pkcs11config)
        self.buffer += '/Contents <'
        self.buffer += self.pkcs11zeros
        self.buffer += '>>>\n'
        self._out('endobj')

    def pkcs11_sign(self):
        sbr = 'SIGNER 0 R'
        dbr = '%-6d 0 R' % self.pkcs11annot
        self.buffer = self.buffer.replace(sbr, dbr, 2)

        buffer = self.buffer.encode('latin1')
        zeros = self.pkcs11zeros.encode('latin1')

        pdfbr1 = buffer.find(zeros)
        pdfbr2 = pdfbr1 + len(zeros)
        br = (0, pdfbr1 - 1, pdfbr2 + 1, len(buffer) - pdfbr2 - 1)
        sbr = b'[0000000000 0000000000 0000000000 0000000000]'
        dbr = b'[%010d %010d %010d %010d]' % br
        buffer = buffer.replace(sbr, dbr, 1)

        b1 = buffer[:br[1]]
        b2 = buffer[br[2]:]
        md = getattr(hashlib, self.pkcs11algomd)()
        md.update(b1)
        md.update(b2)
        signed_md = md.digest()

        contents = signer.sign(None, self.pkcs11key, self.pkcs11cert, self.pkcs11certs, self.pkcs11algomd, True,
                               signed_md)
        contents = self.pkcs11_aligned(contents)

        buffer = buffer.replace(zeros, contents.encode('latin1'), 1)

        self.buffer = buffer.decode('latin1')



import datetime
from endesive import hsm, pdf
from google.cloud import kms
import hashlib

#fill out these variables based on your project
project_id = ""
location_id = ""
key_ring_id = ""
key_id = ""
version_id = ""
filepath = ""

class GoogleHSM(hsm.BaseHSM):
    
    def __init__(self, project_id, location_id, key_ring_id, key_id, version_id):
        
        self.project_id = project_id
        self.location_id = location_id
        self.key_ring_id = key_ring_id
        self.key_id = key_id
        self.version_id = version_id

    def certificate(self):

        """
        See my gist to find out how to create x509 certificates for
        Google HSM-hosted keys: https://gist.github.com/Arbitrage0/de4e0defb20bc539d6db27e4334e0e67
        """

        cert = open('path/to/certificate.crt.pem', 'rb').read()
        return 1, cert

    def sign(self, keyid, data, mech):
       
        """ 
        Following the example here: 
        https://github.com/googleapis/python-kms/blob/master/samples/snippets/sign_asymmetric.py 
        """
        
        client = kms.KeyManagementServiceClient()
        key_version_name = client.crypto_key_version_path(
            self.project_id, 
            self.location_id, 
            self.key_ring_id, 
            self.key_id, 
            self.version_id
        )
        hash_ = getattr(hashlib, mech.lower())(data).digest()
        digest = {mech.lower(): hash_}
        sign_response = client.asymmetric_sign(request={'name': key_version_name, 'digest': digest})
        return sign_response.signature

def main(project_id, location_id, key_ring_id, key_id, version_id, fname):
    
    date = datetime.datetime.utcnow() - datetime.timedelta(hours=12)
    date = date.strftime('D:%Y%m%d%H%M%S+00\'00\'')
    dct = {
        'sigflags': 3,
        'contact': 'user@example.com',
        'location': 'England',
        'signingdate': date.encode(),
        'reason': 'Test',
    }
    Ghsm = GoogleHSM(project_id, location_id, key_ring_id, key_id, version_id)
    datau = open(fname, 'rb').read()
    datas = pdf.cms.sign(datau, dct,
        None, None,
        [],
        'sha256',
        Ghsm,
    )
    fname = fname.replace('.pdf', '-signed-cms-hsm-Google.pdf')
    with open(fname, 'wb') as fp:
        fp.write(datau)
        fp.write(datas)

main(project_id, location_id, key_ring_id, key_id, version_id, filepath)



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
from lxml import etree
import signxml

cert = open("ca/demo2_user1.crt.pem").read()
key = open("ca/demo2_user1.key.pem").read()
data = open("xml.xml", "rb").read()
root = etree.fromstring(data)
signed_root = signxml.XMLSigner(method=signxml.methods.enveloped).sign(
    root, key=key, cert=cert, passphrase=b"1234"
)
verified_data = (
    signxml.XMLVerifier().verify(signed_root, ca_pem_file="ca/demo2_ca.crt.pem").signed_xml
)

xml = etree.tostring(
    signed_root, encoding="UTF-8", xml_declaration=True, standalone=False
)
open("xml-xmlsigner-enveloped.xml", "wb").write(xml)

xml = etree.tostring(
    signed_root,
    encoding="UTF-8",
    xml_declaration=True,
    standalone=False,
    pretty_print=True,
)
open("xml-xmlsigner-enveloped1.xml", "wb").write(xml)



#!/usr/bin/env python3
import sys
from endesive.pdf import PDFVerifier


def main():
    trustedcerts = []
    with open("nccert2016.crt", "rb") as fp:
        trustedcerts.append(fp.read())
    with open("ca/demo2_ca.crt.pem", "rb") as fp:
        trustedcerts.append(fp.read())

    if len(sys.argv) > 1:
        fname = sys.argv[1]
    else:
        fname = "pdf-signed-cms-m32-unizeto.pdf"
    with open(fname, "rb") as fp:
        pdf_data = fp.read()

    v = PDFVerifier(pdf_data, trustedcerts, "/etc/ssl/certs")

    if not v.is_valid_pdf():
        print(f"{fname} is not a pdf file")
        return

    if not v.is_signed():
        if v.modified:
            print(f"file {fname} is modified")
        else:
            print(f"file {fname} is unsigned")
        if not v.whole_file:
            print("the signature does not cover the entire pdf file")
        return
    
    (signed_data, tspdata, crldata, cert, othercerts, hashok, signatureok) = v.decompose_signature()
    if not hashok or not signatureok:
        print(f"file {fname} is modified")
        return

    certok = v.validate_certificate(cert, othercerts)
    if not certok:
        print(f"signing certificate is invalid")
        return

    print(f"signed with cert serial_number: {cert['tbs_certificate']['serial_number'].native}")
    print("other certificates:")
    for ocert in othercerts:
        print(f"     serial_number: {ocert['tbs_certificate']['serial_number'].native}")

    if crldata.native is not None:
        ok, info = v.verify_ocsp_data(cert, othercerts, crldata)
        if ok:
            print(f'ocsp issued at: {info[0]}, next check at: {info[1]}')
        else:
            print(f'ocsp is invalid')

    if tspdata is not None:
        ok, info = v.verify_tsp_data(signed_data, tspdata, othercerts)
        if ok:
            print(f'tsp issued at: {info}')
        else:
            print(f'tsp is invalid')
        
if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
    Signature Annotations
    ~~~~~~~~~~~~~~~~~~~~~~~
    Annotations defined by a width and a height: Square, Circle

    :copyright: Copyright 2019 Autodesk, Inc.
    :license: MIT, see LICENSE for details.
"""
import os.path

from ..pdfrw import PdfArray, PdfName, IndirectPdfDict, PdfDict
from ..pdfttf import TTFFont

from .image import Image
from .text import get_text_commands
from .base import Annotation
from .base import make_border_dict
from ..util.geometry import transform_rect
from ..util.geometry import identity
from ..config.appearance import set_appearance_state
from ..config.appearance import stroke_or_fill
from ..config.constants import DEFAULT_BASE_FONT
from ..config.constants import GRAPHICS_STATE_NAME
from ..config.constants import PDF_ANNOTATOR_FONT
from ..graphics import BaseCommand, FloatTupleCommand
from ..graphics import Bezier
from ..graphics import Close
from ..graphics import ContentStream
from ..graphics import Line
from ..graphics import Move
from ..graphics import quadratic_to_cubic_bezier
from ..graphics import Rect
from ..graphics import Restore
from ..graphics import Save
from ..graphics import BeginText, Text, EndText, TextMatrix
from ..graphics import Font
from ..graphics import FillColor, StrokeColor, Stroke, StrokeWidth, Fill
from ..graphics import CTM, XObject

HELVETICA_PATH = os.path.join(os.path.dirname(__file__), "..", "fonts", "Helvetica.ttf")

class Signature(Annotation):
    """Signatur annotation that defines its location on the document with a
    width and a height.  Internal structure follows that which is documented
    by Adobe, with /frm referencing a blank /n0 layer with all the appearance
    in the stream of the /n2 layer.
    """

    subtype = 'Widget'

    def __init__(self, location, appearance, metadata=None):
        super(Signature, self).__init__(location, appearance, metadata)
        self._images = {}
        self._ttf = {}
        self._type1 = {}

    def add_additional_resources(self, resources):
        # TODO: additional resources like: fonts, ... that need to be included in pdf
        return
        try:
            resources.Font
        except AttributeError:
            resources.Font = IndirectPdfDict()
        for font_name, font in self.ttf.items():
            resources.Font[PdfName(font_name)] = font

    def simple_signature(self, signature):
        """
            simple_signature(self, signature)

            A simple signature annotation whose appearance is inspired by
            the signatures from Adobe's appearances.  The text is output
            using the default font, and no calls to add_image or add_font
            are needed to use this.

            signature = dict(
                background = Image with alpha / None,
                icon = Image with alpha / None,
                labels = bool,
                CN = str,
                DN = str,
                date = str,
                contact = str,
                reason = str,
                location = str,
                software = str,
                outline = [R, G, B],
                border = int,
                )
        """
        bbox = self._internal_location()
        processor = SignatureAppearance(bbox, self._appearance, self._ttf, self._images)
        height = bbox[3] - bbox[1]
        width = bbox[2] - bbox[0]
        t_left = 5
        t_top = 5
        border = signature.get('border', 0)

        if signature.get('labels', False):
            labels = dict(
                DN = 'DN: ',
                CN = 'Digitally signed by ',
                date = 'Date: ',
                contact = 'Contact: ',
                reason = 'Reason: ',
                location = 'Location: ',
                software = 'Signing software: ',
                )
        else:
            labels = {}

        block = []
        for i in ('CN', 'DN', 'date', 'reason', 'contact', 'location', 'software'):
            if i not in signature:
                continue
            block.append('{}{}'.format(labels.get(i, ''), signature[i]))

        cs = ContentStream([Save()])
        if 'background' in signature:
            if type(signature['background']) == list:
                cs.extend(processor.fill_colour(*signature['background']))
                cs.extend(processor.rect_fill(*bbox))
            else:
                if 'bg' not in self._images:
                    self.add_image(signature['background'], 'Bg')
                cs.extend(processor.image('Bg', *bbox))

        cs.extend(processor.fill_colour(*signature.get('outline', [0, 0, 0])))
        cs.extend(processor.stroke_colour(*signature.get('outline', [0, 0, 0])))
        if border:
            cs.extend(processor.border(border))
        if 'icon' in signature:
            if 'icon' not in self._images:
                self.add_image(signature['icon'], 'Icon')
            i_w = self._images['Icon']['/Width']
            i_h = self._images['Icon']['/Height']

            if block:
                factor = 2.0
            else:
                factor = 1.0

            if i_h <= height and i_w <= width/factor:
                # image smaller than alloted half of box
                # vertically centre icon and expand text region
                t_left += i_w + 5
                dist = (height - i_h)/2
                if block:
                    i_box = [bbox[0], bbox[1]+dist, i_w+bbox[0], bbox[3]-dist]
                else:
                    dist2 = (width - i_w)/2
                    i_box = [bbox[0]+dist2, bbox[1]+dist, bbox[2]-dist2, bbox[3]-dist]
            elif (i_w/i_h) > ((width/factor)/height):
                # too wide
                t_left += (width/factor) + 5
                scale = (width/factor)/i_w
                dist = (height - i_h*scale)/2
                i_box = [bbox[0], bbox[1]+dist, width/factor, bbox[3]-dist]
            else:
                # too tall
                scale = height/i_h
                t_left += (i_w*scale) + 5
                if block:
                    i_box = [bbox[0], bbox[1], bbox[0]+i_w*scale, bbox[3]]
                else:
                    # centre if only the icon
                    dist = (width - i_w*scale)/2
                    i_box = [bbox[0]+dist, bbox[1], bbox[2]-dist, bbox[3]]
            cs.extend(processor.image('Icon', *i_box))

        if block:
            font = self.get_default_font()
            cs.extend(processor.text_box('\n'.join(block), font, t_left, t_top, width-2*t_left, height-2*t_top))

        cs.add(Restore())
        self._n2_layer = cs

    def set_signature_appearance(self, *directives):
        """
            set_signature_apperance(self, *directives)

            Use the provided list of lists to compute a signature appearance
            using calls to the SignatureAppearance processor.  Note that the
            text_box directive can only be used with a TTF font.
        """
        processor = SignatureAppearance(self._internal_location(), self._appearance, self._ttf, self._images)
        cs = ContentStream([Save()])
        for x in directives:
            if x[0] in processor.template:
                directives = processor.template[x[0]](processor, *x[1:])
                if type(directives) != list:
                    directives = [directives]
                cs.extend(directives)
        cs.add(Restore())
        self._n2_layer = cs

    def make_rect(self):
        stroke_width = self._appearance.stroke_width
        L = self._location
        return [
            L.x1 - stroke_width,
            L.y1 - stroke_width,
            L.x2 + stroke_width,
            L.y2 + stroke_width,
        ]

    def add_default_font(self):
        self.add_ttf_font(HELVETICA_PATH, PDF_ANNOTATOR_FONT)

    def get_default_font(self):
        if PDF_ANNOTATOR_FONT not in self._ttf:
            self.add_ttf_font(HELVETICA_PATH, PDF_ANNOTATOR_FONT)
        return self._ttf[PDF_ANNOTATOR_FONT]

    def add_type1_font(self, base_font, name="Font", encoding='WinAnsiEncoding'):
        type1 = (
            'Helvetica', 'Helvetica-Bold', 'Helvetica-Oblique', 'Helvetica-BoldOblique',
            'Courier', 'Courier-Bold', 'Courier-Oblique', 'Courier-BoldOblique',
            'Times-Roman', 'Times-Bold', 'Times-Italic', 'Times-BoldItalic',
            'Symbol', 'ZapfDingbats',
            )
        if base_font in ('Symbol', 'ZapfDingbats'):
            encoding = base_font
        self._type1[name] = PdfDict(
            Type = PdfName('Font'),
            Subtype = PdfName('Type1'),
            Name = PdfName(name),
            BaseFont = PdfName(base_font),
            Encoding = PdfName(encoding),
            )

    def add_ttf_font(self, path, name="Font"):
        font = TTFFont(path)
        self._ttf[name] = font

    def add_image(self, obj, name="Image"):
        self._images[name] = Image.make_image_xobject(obj)

    def _make_appearance_stream_dict(self, bounding_box, transform):
        self._make_n0()
        self._make_n2()
        self._make_frm()
        self._make_apn()
        return PdfDict(N=self._apn)

    def as_pdf_object(self, transform, page):
        """Return the PdfDict object representing the annotation, that will be
        inserted as is into the PDF document.

        :param list transform: Transformation matrix to transform the coords
            of the annotation from client-specified space to PDF user space.
        :param PdfDict page: The pdfrw page object from the PDF document
        :returns PdfDict: the annotation object to be inserted into the PDF
        """
        bounding_box = transform_rect(self.make_rect(), transform)
        appearance_stream = self._make_appearance_stream_dict(bounding_box, transform)

        obj = PdfDict(
            Type=PdfName("Annot"),
            Subtype=PdfName(self.subtype),
            Rect=bounding_box,
            AP=appearance_stream,
            P=page,
        )

        self._add_metadata(obj, self._metadata)
        self.add_additional_pdf_object_data(obj)
        obj.indirect = True

        return obj

    def _internal_location(self):
        L = self._location

        return (0, 0, L.x2 - L.x1, L.y2 - L.y1)

    def _make_apn(self):
        self._apn = IndirectPdfDict(
            BBox = self._internal_location(),
            Resources = dict(
                ProcSet = PdfArray([PdfName('PDF')]),
                XObject = {'FRM': self._frm},
            ),
            Type = PdfName('XObject'),
            Subtype = PdfName('Form'),
            FormType = 1,
            Matrix = identity(),
        )
        self._apn['stream'] = 'q 1 0 0 1 0 0 cm /FRM Do Q'

    def _make_frm(self):
        self._frm = IndirectPdfDict(
            BBox = self._internal_location(),
            Resources = dict(
                ProcSet = PdfArray([PdfName('PDF')]),
                XObject = {'n0': self._n0, 'n2': self._n2},
            ),
            Matrix = identity(),
            Type = PdfName('XObject'),
            Subtype = PdfName('Form'),
        )
        self._frm['stream'] = 'q 1 0 0 1 0 0 cm /n0 Do Q q 1 0 0 1 0 0 cm /n2 Do Q'

    def _make_n0(self):
        self._n0 = IndirectPdfDict(
            BBox = self._internal_location(),
            Type = PdfName('XObject'),
            Subtype = PdfName('Form'),
            FormType = 1,
            Matrix = identity(),
            Resources = {'ProcSet': PdfArray([PdfName('Text')])},
        )
        self._n0['stream'] = '% DSBlank'

    def _make_n2(self):
        resources = dict(
                ProcSet = PdfArray([PdfName('PDF'), PdfName('Text'), PdfName('ImageC')])
        )
        fonts = PdfDict()
        if self._ttf:
            for name, font in self._ttf.items():
                fonts[name] = font.get_font()
        if self._type1:
            for name, font in self._type1.items():
                fonts[name] = font
        if fonts:
            resources[PdfName('Font')] = fonts
        if self._images:
            resources[PdfName('XObject')] = self._images
        self._n2 = IndirectPdfDict(
            BBox = self._internal_location(),
            Matrix = identity(),
            Type = PdfName('XObject'),
            Subtype = PdfName('Form'),
            FormType = 1,
            Resources = resources,
            )
        self._n2['stream'] = self._n2_layer.resolve()

class SignatureAppearance():
    def __init__(self, box, appearance, ttf, images):
        self._in_text = False
        self._reset_font = True
        self._reset_tm = False
        self._cur_font = (PDF_ANNOTATOR_FONT, appearance.font_size)
        self._cur_tm = [1, 0, 0, 1, 0, 0]
        self._sc = [0, 0, 0]
        self._fc = [0, 0, 0]
        self._bounds = box
        self._ttf = ttf
        self._images = images

    def fill_colour(self, *colour):
        self._fc = colour
        return [FillColor(*colour)]

    def stroke_colour(self, *colour):
        self._sc = colour
        return [StrokeColor(*colour)]

    def border(self, inset):
        box = self._bounds
        return [
            Rect(inset, inset, box[2]-2*inset, box[3]-2*inset),
            Stroke()
            ]

    def image(self, image_name, x1, y1, x2, y2, distort=True, centred=True):
        if distort:
            scale_x = x2-x1
            scale_y = y2-y1
        else:
            max_w = x2-x1
            max_h = y2-y1
            i_w = self._images[image_name]['/Width']
            i_h = self._images[image_name]['/Height']

            if i_w/i_h >= max_w/max_h:
                # image proportionally wider than box
                scale_x = x2-x1
                scale_y = scale_x*(i_h/i_w)
                if centred: # centre vertically
                    dist = (max_h-scale_y)/2.0
                    y1 += dist
            elif i_w/i_h < max_w/max_h:
                # image proportionally narrower than box
                scale_y = y2-y1
                scale_x = scale_y*(i_w/i_h)
                if centred: # centre horizontally
                    dist = (max_w-scale_x)/2.0
                    x1 += dist

        commands = []
        if self._in_text:
            commands.append(EndText())
            self._reset_font = True
            self._reset_tm = True
            self._in_text = False
        commands.append(Save())
        commands.append(CTM((scale_x, 0, 0, scale_y, x1, y1)))
        commands.append(XObject(image_name))
        commands.append(Restore())
        return commands

    def rect(self, *box):
        seq = []
        if self._in_text:
            seq.append(EndText())
            self._in_text = False
        seq.extend([
            Rect(*box),
            Stroke()
            ])
        return seq

    def rect_fill(self, *box):
        seq = []
        if self._in_text:
            seq.append(EndText())
            self._in_text = False
        seq.extend([
            Rect(*box),
            Fill()
            ])
        return seq

    def reset(self):
        return [
            StrokeColor(0, 0, 0),
            FillColor(0, 0, 0),
            ]

    def text_position(self, x, y):
        if x < 0:
            x = self._bounds[2]+x
        if y < 0:
            y = self._bounds[3]+y

        self._cur_tm[4] = x
        self._cur_tm[5] = y
        if self._in_text:
            return [TextMatrix(self._cur_tm.copy())]
        self._reset_tm = True
        return []

    def font(self, name, size):
        if name == 'default':
            name = PDF_ANNOTATOR_FONT

        self._cur_font = (name, size)
        if self._in_text:
            return [Font(name, size), TextLeading(size*1.2)]
        self._reset_font = True
        return []

    def text_box(self, text, ttf_font, x, y, width, height, font_size=8, wrap_text=True, align='left', baseline='middle', line_spacing=1.2):
        if type(ttf_font) == str:
            font = self._ttf.get(ttf_font, None)
            if font is None:
                font = self._ttf[PDF_ANNOTATOR_FONT]
                fontname = PDF_ANNOTATOR_FONT
            else:
                fontname = ttf_font
        else:
            font = ttf_font
            fontname = font.font["name"]
            if fontname == 'Helvetica':
                fontname = PDF_ANNOTATOR_FONT
        font.set_size(font_size)
        font.set_text(text)

        commands = []
        if not self._in_text:
            commands.extend([BeginText(), Font(fontname, font_size)])
        commands.extend(get_text_commands(x, y, x+width, y+height, text, font_size, wrap_text, align, baseline, line_spacing, font))
        if not self._in_text:
            commands.append(EndText())
        return commands

    def text(self, text):
        if self._cur_font[0] in self._ttf:
            ttf_font = self._ttf[self._cur_font[0]]
            ttf_font.set_size(self._cur_font[1])
            ttf_font.set_text(text)
            text = text.encode("utf-16be").decode("latin1")
        commands = []
        if not self._in_text:
            commands.append(BeginText())
            self._in_text = True
        if self._reset_tm:
            commands.append(TextMatrix(self._cur_tm.copy()))
            self._reset_tm = False
        if self._reset_font:
            commands.append(Font(*self._cur_font))
            commands.append(TextLeading(self._cur_font[1]*1.2))
            self._reset_font = False
        commands.append(Text(text))
        return commands

    def new_line(self):
        if self._in_text:
            return [ NewLine() ]
        return []

    def done(self):
        if self._in_text:
            self._in_text = False
            return [ EndText() ]
        return []

    template = dict(
        save = Save,
        reset = reset,
        image = image,
        rect = rect,
        rect_fill = rect_fill,
        border = border,
        font = font,
        text = text,
        text_box = text_box,
        text_position = text_position,
        new_line = new_line,
        done = done,
        fill_colour = fill_colour,
        stroke_colour = stroke_colour,
        fill_color = fill_colour,
        stroke_color = stroke_colour,
    )

class TextPosition(metaclass=FloatTupleCommand):
    COMMAND = 'Td'
    ARGS = ['x', 'y']

class TextRender(metaclass=FloatTupleCommand):
    COMMAND = 'Tr'
    ARGS = ['render']

class TextScale(metaclass=FloatTupleCommand):
    COMMAND = 'Tz'
    ARGS = ['scale']

class TextLeading(metaclass=FloatTupleCommand):
    COMMAND = 'TL'
    ARGS = ['leading']

class TextRise(metaclass=FloatTupleCommand):
    COMMAND = 'Ts'
    ARGS = ['rise']

class WordSpacing(metaclass=FloatTupleCommand):
    COMMAND = 'Tw'
    ARGS = ['wordSpace']

class CharacterSpacing(metaclass=FloatTupleCommand):
    COMMAND = 'Tc'
    ARGS = ['charSpace']

class StrokeGray(metaclass=FloatTupleCommand):
    COMMAND = 'G'
    ARGS = ['gray']

class FillGray(metaclass=FloatTupleCommand):
    COMMAND = 'g'
    ARGS = ['gray']

class NewLine(BaseCommand):
    COMMAND = 'T*'



# -*- coding: utf-8 -*-
"""
    GraphicsState
    ~~~~~~~~~~~~~
    Configuration for an annotation's graphics state.

    :copyright: Copyright 2019 Autodesk, Inc.
    :license: MIT, see LICENSE for details.
"""
import attr
from ..pdfrw import PdfDict, PdfName

from ..config.constants import ALLOWED_LINE_CAPS
from ..config.constants import ALLOWED_LINE_JOINS
from ..util.validation import between
from ..util.validation import Enum
from ..util.validation import Field
from ..util.validation import Number
from ..util.validation import positive
from ..util.validation import validate_dash_array


NAME_TO_PDF_ATTR = {
    "line_width": "LW",
    "line_cap": "LC",
    "line_join": "LJ",
    "miter_limit": "ML",
    "dash_array": "D",
    "stroke_transparency": "CA",
    "fill_transparency": "ca",
}


@attr.s
class GraphicsState(object):
    """External graphics state config object, that can be used with explicit
    content streams to control annotation appearance.

    Some of these values can also be specified by their own operators in the
    content stream. For example, the line_width property can also be specified
    by the StrokeWidth (w) content stream operator.

    See the full PDF spec for constraints on and descriptions of these values.
    There are a lot more graphics state options, but they are highly technical
    and beyond the scope of this library.
    """

    line_width = Number(default=None, validator=positive)
    line_cap = Enum(ALLOWED_LINE_CAPS, default=None)
    line_join = Enum(ALLOWED_LINE_JOINS, default=None)
    miter_limit = Number(default=None)
    dash_array = Field(list, validator=validate_dash_array, default=None)
    stroke_transparency = Number(default=None, validator=between(0, 1))
    fill_transparency = Number(default=None, validator=between(0, 1))

    def as_pdf_dict(self):
        pdf_dict = PdfDict(Type=PdfName("ExtGState"))
        for attr_name, attr_value in self.__dict__.items():
            if attr_value is not None:
                pdf_name = NAME_TO_PDF_ATTR[attr_name]
                pdf_dict[PdfName(pdf_name)] = attr_value
        return pdf_dict

    def has_content(self):
        """Returns True if any of the attributes is non-null."""
        return any(value is not None for value in self.__dict__.values())



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import typing
import datetime
import os, os.path
import sys
import glob
import uuid

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization import pkcs12
from cryptography.x509.oid import NameOID

force = "--force" in sys.argv

(
    ca,
    ca_key,
    cert1,
    cert1_key,
    cert1_pub,
    cert1_p12,
    cert2,
    cert2_key,
    cert2_pub,
    cert2_p12,
    cert3,
    cert3_key,
    cert3_pub,
    cert3_p12,
) = (
    "demo2_ca.crt.pem",
    "demo2_ca.key.pem",
    "demo2_user1.crt.pem",
    "demo2_user1.key.pem",
    "demo2_user1.pub.pem",
    "demo2_user1.p12",
    "demo2_user2.crt.pem",
    "demo2_user2.key.pem",
    "demo2_user2.pub.pem",
    "demo2_user2.p12",
    "demo2_user3.crt.pem",
    "demo2_user3.key.pem",
    "demo2_user3.pub.pem",
    "demo2_user3.p12",
)


class Main(object):
    def key_create(self) -> rsa.RSAPrivateKey:
        return rsa.generate_private_key(
            public_exponent=65537, key_size=2048, backend=default_backend()
        )

    def key_save(self, fname: str, key: rsa.RSAPrivateKey, password: str) -> None:
        # Write our key to disk for safe keeping
        with open(os.path.join("ca", fname), "wb") as f:
            if not password:
                f.write(
                    key.public_key().public_bytes(
                        encoding=serialization.Encoding.PEM,
                        format=serialization.PublicFormat.SubjectPublicKeyInfo,
                    )
                )
            else:
                f.write(
                    key.private_bytes(
                        encoding=serialization.Encoding.PEM,
                        format=serialization.PrivateFormat.TraditionalOpenSSL,
                        encryption_algorithm=serialization.BestAvailableEncryption(
                            password.encode("utf-8")
                        ),
                    )
                )

    def key_load(self, fname: str, password: str) -> rsa.RSAPrivateKey:
        with open(os.path.join("ca", fname), "rb") as f:
            private_key = serialization.load_pem_private_key(
                f.read(), password.encode("utf-8"), default_backend()
            )
            return private_key

    def cert_save(self, fname: str, data: x509.Certificate) -> None:
        with open(os.path.join("ca", fname), "wb") as f:
            f.write(data.public_bytes(serialization.Encoding.PEM))
        with open(os.path.join("ca", fname) + ".cer", "wb") as f:
            f.write(data.public_bytes(serialization.Encoding.DER))
        cwd = os.getcwd()
        os.chdir("ca")
        os.symlink(fname, str(data.serial_number))
        os.chdir(cwd)

    def cert_load(self, fname: str) -> x509.Certificate:
        with open(os.path.join("ca", fname), "rb") as f:
            return x509.load_pem_x509_certificate(f.read(), default_backend())

    def csr_load(self, fname: str) -> x509.CertificateSigningRequest:
        with open(os.path.join("ca", fname), "rb") as f:
            return x509.load_pem_x509_csr(data=f.read(), backend=default_backend())

    def csr_create(
        self,
        email: str,
        key: rsa.RSAPrivateKey,
        country: typing.Union[str, None] = None,
        state: typing.Union[str, None] = None,
        locality: typing.Union[str, None] = None,
        organization: typing.Union[str, None] = None,
        commonname: typing.Union[str, None] = None,
    ) -> x509.CertificateSigningRequest:
        names = []
        for t, v in (
            (NameOID.COUNTRY_NAME, country),
            (NameOID.STATE_OR_PROVINCE_NAME, state),
            (NameOID.LOCALITY_NAME, locality),
            (NameOID.ORGANIZATION_NAME, organization),
            (NameOID.COMMON_NAME, commonname),
        ):
            if v:
                names.append(x509.NameAttribute(t, v))
        names.append(x509.NameAttribute(NameOID.EMAIL_ADDRESS, email))
        return (
            x509.CertificateSigningRequestBuilder()
            .subject_name(x509.Name(names))
            .sign(
                # Sign the CSR with our private key.
                key,
                hashes.SHA256(),
                default_backend(),
            )
        )

    def csr_sign(self, csr: x509.CertificateSigningRequest) -> x509.Certificate:
        emails = csr.subject.get_attributes_for_oid(NameOID.EMAIL_ADDRESS)
        names = [x509.RFC822Name(emails[0].value)]
        return (
            x509.CertificateBuilder()
            .subject_name(csr.subject)
            .issuer_name(self.ca_cert.subject)
            .public_key(csr.public_key())
            .serial_number(uuid.uuid4().int)  # pylint: disable=no-member
            .not_valid_before(datetime.datetime.utcnow())
            .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=365))
            .add_extension(
                x509.BasicConstraints(ca=False, path_length=None), critical=True
            )
            .add_extension(
                x509.AuthorityKeyIdentifier.from_issuer_public_key(
                    self.ca_pk.public_key()
                ),
                critical=False,
            )
            .add_extension(
                x509.CRLDistributionPoints(
                    [
                        x509.DistributionPoint(
                            full_name=[
                                x509.UniformResourceIdentifier(
                                    "http://ca.trisoft.com.pl/crl"
                                )
                            ],
                            relative_name=None,
                            reasons=None,
                            crl_issuer=None,
                        )
                    ]
                ),
                critical=False,
            )
            .add_extension(
                x509.AuthorityInformationAccess(
                    [
                        x509.AccessDescription(
                            x509.OID_CA_ISSUERS,
                            x509.UniformResourceIdentifier(
                                "http://ca.trisoft.com.pl/cacert"
                            ),
                        ),
                        x509.AccessDescription(
                            x509.OID_OCSP,
                            x509.UniformResourceIdentifier(
                                "http://ca.trisoft.com.pl/ocsp"
                            ),
                        ),
                    ]
                ),
                critical=False,
            )
            .add_extension(
                x509.SubjectAlternativeName(names),
                critical=False,
            )
            # certificate_policies
            .add_extension(
                x509.ExtendedKeyUsage(
                    [x509.OID_CLIENT_AUTH, x509.OID_EMAIL_PROTECTION]
                ),
                critical=False,
            )
            .add_extension(
                x509.SubjectKeyIdentifier.from_public_key(csr.public_key()),
                critical=False,
            )
            .add_extension(
                x509.KeyUsage(
                    digital_signature=True,
                    content_commitment=True,  # nonRepudiation
                    key_encipherment=True,
                    data_encipherment=False,
                    key_agreement=True,
                    encipher_only=False,
                    decipher_only=False,
                    # ca
                    key_cert_sign=False,
                    crl_sign=False,
                ),
                critical=True,
            )
            .sign(
                private_key=self.ca_pk,
                algorithm=hashes.SHA256(),
                backend=default_backend(),
            )
        )

    def pk12_save(
        self,
        name: bytes,
        cert: x509.Certificate,
        key: rsa.RSAPrivateKey,
        fname: str,
        password: str,
    ) -> None:
        data = pkcs12.serialize_key_and_certificates(
            name=name,
            key=key,
            cert=cert,
            cas=[self.ca_cert],
            encryption_algorithm=serialization.BestAvailableEncryption(
                password.encode("utf8")
            ),
        )
        with open(os.path.join("ca", fname), "wb") as f:
            f.write(data)

    def pk12_load(self, fname: str, password: str) -> pkcs12.PKCS12KeyAndCertificates:
        with open(os.path.join("ca", fname), "rb") as fp:
            return pkcs12.load_key_and_certificates(
                fp.read(), password.encode("utf-8"), default_backend()
            )

    def ca_create(self, key: rsa.RSAPrivateKey) -> x509.Certificate:
        subject = issuer = x509.Name(
            [
                x509.NameAttribute(NameOID.COMMON_NAME, "CA"),
            ]
        )
        return (
            x509.CertificateBuilder()
            .subject_name(subject)
            .issuer_name(issuer)
            .public_key(key.public_key())
            .serial_number(x509.random_serial_number())
            .not_valid_before(datetime.datetime.utcnow())
            .not_valid_after(
                # Our certificate will be valid for 10 years
                datetime.datetime.utcnow()
                + datetime.timedelta(days=10 * 365)
            )
            .add_extension(
                x509.BasicConstraints(
                    ca=True,
                    path_length=None,  # pathlen: is equal to the number of CAs/ICAs it can sign
                ),
                critical=True,
            )
            .add_extension(
                x509.AuthorityKeyIdentifier.from_issuer_public_key(key.public_key()),
                critical=False,
            )
            .add_extension(
                x509.SubjectKeyIdentifier.from_public_key(key.public_key()),
                critical=False,
            )
            .add_extension(
                x509.KeyUsage(
                    digital_signature=True,
                    content_commitment=False,  # nonRepudiation
                    key_encipherment=False,
                    data_encipherment=False,
                    key_agreement=False,
                    encipher_only=False,
                    decipher_only=False,
                    # ca
                    key_cert_sign=True,
                    crl_sign=True,
                ),
                critical=True,
            )
            .sign(
                # Sign our certificate with our private key
                key,
                hashes.SHA256(),
                default_backend(),
            )
        )

    def CA(self) -> None:
        create = force or not (
            os.path.exists(os.path.join("ca", ca))
            and os.path.exists(os.path.join("ca", ca_key))
        )
        if create:
            for fqname in glob.glob("ca/*"):
                os.unlink(fqname)
            print("CA generating certificate")
            ca_pk = self.key_create()
            ca_cert = self.ca_create(ca_pk)

            self.key_save(ca_key, ca_pk, "1234")
            self.cert_save(ca, ca_cert)
        else:
            print("CA using certificate")
            ca_pk = self.key_load(ca_key, "1234")
            ca_cert = self.cert_load(ca)

        self.ca_cert = ca_cert
        self.ca_pk = ca_pk

    def USER(self, no, cert, cert_key, cert_pub, cert_p12) -> None:
        create = force or not (
            os.path.exists(os.path.join("ca", cert))
            and os.path.exists(os.path.join("ca", cert_key))
            and os.path.exists(os.path.join("ca", cert_pub))
            and os.path.exists(os.path.join("ca", cert_p12))
        )
        if create:
            client_pk = self.key_create()
            client_csr = self.csr_create("demo%d@trisoft.com.pl" % no, client_pk)
            client_cert = self.csr_sign(client_csr)
            self.cert_save(cert, client_cert)
            self.key_save(cert_key, client_pk, "1234")
            self.key_save(cert_pub, client_pk, None)
            self.pk12_save(
                "USER cert".encode("utf-8"), client_cert, client_pk, cert_p12, "1234"
            )

            # os.chmod(pkcs12 % user, stat.S_IRUSR | stat.S_IWUSR)  # 0o600 perms

    def USERs(self):
        self.USER(1, cert1, cert1_key, cert1_pub, cert1_p12)
        self.USER(2, cert2, cert2_key, cert2_pub, cert2_p12)
        self.USER(3, cert3, cert3_key, cert3_pub, cert3_p12)


print("Generating certificates")
cls = Main()
cls.CA()
cls.USERs()



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
from cryptography.hazmat import backends
from cryptography.hazmat.primitives.serialization import pkcs12
from endesive import email

import hashlib
from asn1crypto import cms, algos, core, pem, x509

def main():
    with open('ca/demo2_user1.p12', 'rb') as fp:
        p12 = pkcs12.load_key_and_certificates(fp.read(), b'1234', backends.default_backend())
    datau = open('smime-unsigned.txt', 'rb').read()

    datau1 = datau.replace(b'\n', b'\r\n')
    hashalgo = 'sha256'
    signed_value = getattr(hashlib, hashalgo)(datau1).digest()
    attrs = [
        cms.CMSAttribute({
            'type': cms.CMSAttributeType('content_type'),
            'values': ('data',),
        }),
        cms.CMSAttribute({
            'type': cms.CMSAttributeType('message_digest'),
            'values': (signed_value,),
        }),
    ]

    datas = email.sign(datau,
        p12[0], p12[1], p12[2],
        'sha256',
        attrs=attrs
    )
    open('smime-signed-attr-custom.txt', 'wb').write(datas)


main()



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import sys
import pprint
import binascii
from cryptography.hazmat import backends
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.serialization import pkcs12
from asn1crypto import x509, pem


def cert2asn(cert):
    cert_bytes = cert.public_bytes(serialization.Encoding.PEM)
    if pem.detect(cert_bytes):
        _, _, cert_bytes = pem.unarmor(cert_bytes)
    return x509.Certificate.load(cert_bytes)

def main():
    if len(sys.argv) == 3:
        p12name = sys.argv[1]
        p12pass = sys.argv[2]
    else:
        p12name = 'ca/demo2_user1.p12'
        p12pass = '1234'
    print(len(sys.argv), p12name, p12pass)
    with open(p12name, 'rb') as fp:
        p12pk, p12pc, p12oc = pkcs12.load_key_and_certificates(fp.read(), p12pass.encode(), backends.default_backend())
    signature = p12pk.sign(
        b"message",
        padding.PKCS1v15(),
        hashes.SHA1()
    )
    cert = cert2asn(p12pc)
    print('issuer', cert.issuer.native)
    print('subject', cert.subject.native)
    pprint.pprint(cert.native)
main()



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import sys
from endesive import pdf, hsm

import os
import sysconfig

os.environ['SOFTHSM2_CONF'] = 'softhsm2.conf'
if not os.path.exists(os.path.join(os.getcwd(), 'softhsm2.conf')):
    open('softhsm2.conf', 'wt').write('''\
log.level = DEBUG
directories.tokendir = %s/softhsm2/
objectstore.backend = file
slots.removable = false
''' % os.getcwd())
if not os.path.exists(os.path.join(os.getcwd(), 'softhsm2')):
    os.mkdir(os.path.join(os.getcwd(), 'softhsm2'))

#
#!/bin/bash
#SOFTHSM2_CONF=softhsm2.conf
#softhsm2-util --label "endesive" --slot 1 --init-token --pin secret1 --so-pin secret2
#softhsm2-util --show-slots
#

if sys.platform == 'win32':
    dllpath = r'W:\binw\SoftHSM2\lib\softhsm2-x64.dll'
else:
    dllpath = os.path.join(sysconfig.get_config_var('LIBDIR'), "softhsm/libsofthsm2.so")

import PyKCS11 as PK11

class Signer(hsm.HSM):
    def certificate(self):
        self.login("endesieve", "secret1")
        keyid = bytes((0x66,0x66,0x90))
        try:
            pk11objects = self.session.findObjects([(PK11.CKA_CLASS, PK11.CKO_CERTIFICATE)])
            all_attributes = [
                #PK11.CKA_SUBJECT,
                PK11.CKA_VALUE,
                #PK11.CKA_ISSUER,
                #PK11.CKA_CERTIFICATE_CATEGORY,
                #PK11.CKA_END_DATE,
                PK11.CKA_ID,
            ]

            for pk11object in pk11objects:
                try:
                    attributes = self.session.getAttributeValue(pk11object, all_attributes)
                except PK11.PyKCS11Error as e:
                    continue

                attrDict = dict(list(zip(all_attributes, attributes)))
                cert = bytes(attrDict[PK11.CKA_VALUE])
                if keyid == bytes(attrDict[PK11.CKA_ID]):
                    return keyid, cert
        finally:
            self.logout()
        return None, None

    def sign(self, keyid, data, mech):
        self.login("endesieve", "secret1")
        try:
            privKey = self.session.findObjects([(PK11.CKA_CLASS, PK11.CKO_PRIVATE_KEY), (PK11.CKA_ID, keyid)])[0]
            mech = getattr(PK11, 'CKM_%s_RSA_PKCS' % mech.upper())
            sig = self.session.sign(privKey, data, PK11.Mechanism(mech, None))
            return bytes(sig)
        finally:
            self.logout()

import binascii
from cryptography.hazmat import backends
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.serialization import pkcs12
from asn1crypto import x509, pem


def cert2asn(cert, cert_bytes=True):
    if isinstance(cert, x509.Certificate):
        return cert
    if cert_bytes:
        cert_bytes = cert.public_bytes(serialization.Encoding.PEM)
    else:
        cert_bytes = cert
    if pem.detect(cert_bytes):
        _, _, cert_bytes = pem.unarmor(cert_bytes)
    return x509.Certificate.load(cert_bytes)

def main():
    hsm = Signer(dllpath)
    hsm.login("endesieve", "secret1")
    keyid = bytes((0x66,0x66,0x90))
    ic = hsm.cert_load(keyid)
    hsm.logout()
    print(ic)
    ic = cert2asn(ic, False)
    print(ic)
main()



#!/usr/bin/env vpython2
import os
import os.path
import smtplib
import datetime

from email.MIMEMultipart import MIMEMultipart
from email.MIMEBase import MIMEBase
from email.MIMEText import MIMEText
from email.Utils import COMMASPACE, formatdate
from email import Encoders

from M2Crypto import BIO, Rand, SMIME, X509

randpool = '/tmp/randpool.dat'

ca_cert = 'ca/demo2_ca.crt.pem'
# we need to have access to both keys
signer_key  = 'ca/demo2_user1.key.pem'
signer_cert = 'ca/demo2_user1.crt.pem'
recipient_key  = 'ca/demo2_user2.key.pem'
recipient_cert = 'ca/demo2_user2.crt.pem'

def makebuf(text):
    return BIO.MemoryBuffer(text)

class Demo:
    def sign(self):

        # Make a MemoryBuffer of the message.
        buf = makebuf('a sign of our times')

        # Seed the PRNG.
        Rand.load_file(randpool, -1)

        # Instantiate an SMIME object; set it up; sign the buffer.
        s = SMIME.SMIME()
        s.load_key(signer_key, signer_cert)

        p7 = s.sign(buf, SMIME.PKCS7_DETACHED)

        # Recreate buf.
        buf = makebuf('a sign of our times')

        # Output p7 in mail-friendly format.
        out = BIO.MemoryBuffer()
        out.write('From: sender@example.dom\n')
        out.write('To: recipient@example.dom\n')
        out.write('Subject: M2Crypto S/MIME testing\n')

        s.write(out, p7, buf)

        result = out.read()

        # Save the PRNG's state.
        Rand.save_file(randpool)

        open('smime-m2-sign.txt', 'wt').write(result)

    def verify(self):

        # Instantiate an SMIME object.
        s = SMIME.SMIME()

        # Load the signer's cert.
        x509 = X509.load_cert(signer_cert)
        sk = X509.X509_Stack()
        sk.push(x509)
        s.set_x509_stack(sk)

        # Load the signer's CA cert.
        st = X509.X509_Store()
        st.load_info(ca_cert)
        s.set_x509_store(st)

        # Load the data, verify it.
        p7, data = SMIME.smime_load_pkcs7('smime-m2-sign.txt')
        v = s.verify(p7)

        print v
        print data
        print data.read()

    def encrypt(self):

        # Make a MemoryBuffer of the message.
        buf = makebuf('a sign of our times')

        # Seed the PRNG.
        Rand.load_file(randpool, -1)

        # Instantiate an SMIME object.
        s = SMIME.SMIME()

        # Load target cert to encrypt to.
        x509 = X509.load_cert(recipient_cert)
        sk = X509.X509_Stack()
        sk.push(x509)
        s.set_x509_stack(sk)

        # Set cipher: 3-key triple-DES in CBC mode.
        s.set_cipher(SMIME.Cipher('des_ede3_cbc'))

        # Encrypt the buffer.
        p7 = s.encrypt(buf)

        # Output p7 in mail-friendly format.
        out = BIO.MemoryBuffer()
        out.write('From: sender@example.dom\n')
        out.write('To: recipient@example.dom\n')
        out.write('Subject: M2Crypto S/MIME testing\n')
        s.write(out, p7)

        result = out.read()

        # Save the PRNG's state.
        Rand.save_file(randpool)

        open('smime-m2-encrypt.txt', 'wt').write(result)

    def decrypt(self):
        # Instantiate an SMIME object.
        s = SMIME.SMIME()

        # Load private key and cert.
        s.load_key(recipient_key, recipient_cert)

        # Load the encrypted data.
        p7, data = SMIME.smime_load_pkcs7('smime-m2-encrypt.txt')

        # Decrypt p7.
        out = s.decrypt(p7)

        print out

    def sign_and_encrypt(self):

        # Make a MemoryBuffer of the message.
        buf = makebuf('a sign of our times')

        # Seed the PRNG.
        Rand.load_file(randpool, -1)

        # Instantiate an SMIME object.
        s = SMIME.SMIME()

        # Load signer's key and cert. Sign the buffer.
        s.load_key(signer_key, signer_cert)
        p7 = s.sign(buf, SMIME.PKCS7_DETACHED)

        # Load target cert to encrypt the signed message to.
        x509 = X509.load_cert(recipient_cert)
        sk = X509.X509_Stack()
        sk.push(x509)
        s.set_x509_stack(sk)

        # Set cipher: 3-key triple-DES in CBC mode.
        s.set_cipher(SMIME.Cipher('des_ede3_cbc'))

        # Create a temporary buffer.
        tmp = BIO.MemoryBuffer()

        # Write the signed message into the temporary buffer.
        s.write(tmp, p7, buf)

        # Encrypt the temporary buffer.
        p7 = s.encrypt(tmp)

        # Output p7 in mail-friendly format.
        out = BIO.MemoryBuffer()
        out.write('From: sender@example.dom\n')
        out.write('To: recipient@example.dom\n')
        out.write('Subject: M2Crypto S/MIME testing\n')
        s.write(out, p7)

        result = out.read()

        # Save the PRNG's state.
        Rand.save_file(randpool)

        open('smime-m2-sign-encrypt.txt', 'wt').write(result)

    def decrypt_and_verify(self):
        # Instantiate an SMIME object.
        s = SMIME.SMIME()

        # Load private key and cert.
        s.load_key(recipient_key, recipient_cert)

        # Load the signed/encrypted data.
        p7, data = SMIME.smime_load_pkcs7('smime-m2-sign-encrypt.txt')

        # After the above step, 'data' == None.
        # Decrypt p7. 'out' now contains a PKCS #7 signed blob.
        out = s.decrypt(p7)

        # Load the signer's cert.
        x509 = X509.load_cert(signer_cert)
        sk = X509.X509_Stack()
        sk.push(x509)
        s.set_x509_stack(sk)

        # Load the signer's CA cert.
        st = X509.X509_Store()
        st.load_info(ca_cert)
        s.set_x509_store(st)

        # Recall 'out' contains a PKCS #7 blob.
        # Transform 'out'; verify the resulting PKCS #7 blob.
        p7_bio = BIO.MemoryBuffer(out)
        p7, data = SMIME.smime_load_pkcs7_bio(p7_bio)
        v = s.verify(p7)

        print v

    def sign_and_attachment(self):
        server = 'mail.example.dom'
        sender = 'sender@example.dom'
        to = ['recipient@example.dom',]
        subject = 'test'
        text = 'test message'
        files=['m2-demo.py']
        attachments={}
        bcc=[]


        if isinstance(to,str):
            to = [to]

        # create multipart message
        msg = MIMEMultipart()

        # attach message text as first attachment
        msg.attach( MIMEText(text) )

        # attach files to be read from file system
        for file in files:
            part = MIMEBase('application', "octet-stream")
            part.set_payload( open(file,"rb").read() )
            Encoders.encode_base64(part)
            part.add_header('Content-Disposition', 'attachment; filename="%s"'
                           % os.path.basename(file))
            msg.attach(part)

        # attach filest read from dictionary
        for name in attachments:
            part = MIMEBase('application', "octet-stream")
            part.set_payload(attachments[name])
            Encoders.encode_base64(part)
            part.add_header('Content-Disposition', 'attachment; filename="%s"' % name)
            msg.attach(part)

        # put message with attachments into into SSL' I/O buffer
        msg_str = msg.as_string()
        buf = BIO.MemoryBuffer(msg_str)

        # load seed file for PRNG
        Rand.load_file(randpool, -1)

        smime = SMIME.SMIME()

        # load certificate
        smime.load_key(signer_key, signer_cert)

        # sign whole message
        p7 = smime.sign(buf, SMIME.PKCS7_DETACHED)

        # create buffer for final mail and write header
        out = BIO.MemoryBuffer()
        out.write('From: %s\n' % sender)
        out.write('To: %s\n' % COMMASPACE.join(to))
        out.write('Date: %s\n' % formatdate(localtime=True))
        out.write('Subject: %s\n' % subject)
        out.write('Auto-Submitted: %s\n' % 'auto-generated')

        # convert message back into string
        buf = BIO.MemoryBuffer(msg_str)

        # append signed message and original message to mail header
        smime.write(out, p7, buf)

        # load save seed file for PRNG
        Rand.save_file(randpool)

        # extend list of recipents with bcc adresses
        to.extend(bcc)

        result = out.read()

        open('smime-m2-attachment.txt', 'wt').write(result)
        return

        # finaly send mail
        smtp = smtplib.SMTP(server)
        smtp.sendmail(sender, to, result )
        smtp.close()

def main():
    cls = Demo()

    if 0:
        cls.sign()
        cls.encrypt()
        cls.sign_and_encrypt()
        cls.sign_and_attachment()

    cls.verify()
    cls.decrypt()
    cls.decrypt_and_verify()

main()



# *-* coding: utf-8 *-*
from __future__ import unicode_literals

import sys
import types
import hashlib
import time
from base64 import b64encode
from datetime import datetime

import requests
import pytz
from asn1crypto import cms, algos, core, keys, pem, tsp, x509, ocsp, util
from oscrypto import asymmetric
from cryptography.hazmat import backends
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, utils, ec


def cert2asn(cert, cert_bytes=True):
    if isinstance(cert, x509.Certificate):
        return cert
    if cert_bytes:
        cert_bytes = cert.public_bytes(serialization.Encoding.PEM)
    else:
        cert_bytes = cert
    if pem.detect(cert_bytes):
        _, _, cert_bytes = pem.unarmor(cert_bytes)
    return x509.Certificate.load(cert_bytes)


def timestamp(unhashed, hashalgo, url, credentials, req_options, prehashed=None):
    if prehashed:
        hashed_value = prehashed
    else:
        hashed_value = getattr(hashlib, hashalgo)(unhashed).digest()
    tspreq = tsp.TimeStampReq(
        {
            "version": 1,
            "message_imprint": tsp.MessageImprint(
                {
                    "hash_algorithm": algos.DigestAlgorithm({"algorithm": hashalgo}),
                    "hashed_message": hashed_value,
                }
            ),
            #'req_policy', ObjectIdentifier, {'optional': True}),
            "nonce": int(time.time() * 1000),
            "cert_req": True,
            #'extensions': tsp.Extensions()
        }
    )
    tspreq = tspreq.dump()

    tspheaders = {"Content-Type": "application/timestamp-query"}
    if credentials is not None:
        username = credentials.get("username", None)
        password = credentials.get("password", None)
        if username and password:
            auth_header_value = b64encode(
                bytes(username + ":" + password, "utf-8")
            ).decode("ascii")
            tspheaders["Authorization"] = f"Basic {auth_header_value}"
    if req_options is None:
        req_options = {}

    tspresp = requests.post(url, data=tspreq, headers=tspheaders, **req_options)
    if tspresp.headers.get("Content-Type", None) == "application/timestamp-reply":
        tspresp = tsp.TimeStampResp.load(tspresp.content)

        if tspresp["status"]["status"].native == "granted":
            attrs = [
                cms.CMSAttribute(
                    {
                        "type": cms.CMSAttributeType("signature_time_stamp_token"),
                        "values": cms.SetOfContentInfo(
                            [
                                cms.ContentInfo(
                                    {
                                        "content_type": cms.ContentType("signed_data"),
                                        "content": tspresp["time_stamp_token"][
                                            "content"
                                        ],
                                    }
                                )
                            ]
                        ),
                    }
                )
            ]
            return attrs
        else:
            raise ValueError("TimeStampResponse status is not granted")
    else:
        raise ValueError("TimeStampResponse has invalid content type")


def sign(
    datau,
    key,
    cert,
    othercerts,
    hashalgo,
    attrs=True,
    signed_value=None,
    hsm=None,
    pss=False,
    timestampurl=None,
    timestampcredentials=None,
    timestamp_req_options=None,
    ocspurl=None,
    ocspissuer=None,
):
    if signed_value is None:
        signed_value = getattr(hashlib, hashalgo)(datau).digest()
    signed_time = datetime.now(tz=util.timezone.utc)

    if hsm is not None:
        keyid, cert = hsm.certificate()
        cert = cert2asn(cert, False)
    else:
        cert = cert2asn(cert)

    certissuer = None
    certificates = []
    certificates.append(cert)
    for i in range(len(othercerts)):
        certo = cert2asn(othercerts[i])
        if certo.subject == cert.issuer:
            certissuer = certo
        certificates.append(certo)

    hashalgo = unicode(hashalgo) if sys.version[0] < "3" else hashalgo

    signer = {
        "version": "v1",
        "sid": cms.SignerIdentifier(
            {
                "issuer_and_serial_number": cms.IssuerAndSerialNumber(
                    {
                        "issuer": cert.issuer,
                        "serial_number": cert.serial_number,
                    }
                ),
            }
        ),
        "digest_algorithm": algos.DigestAlgorithm({"algorithm": hashalgo}),
        "signature": signed_value,
    }
    if not pss:
        signer["signature_algorithm"] = algos.SignedDigestAlgorithm(
            {"algorithm": "rsassa_pkcs1v15"}
        )
    else:
        md = getattr(hashes, hashalgo.upper())
        if isinstance(key, keys.PrivateKeyInfo):
            salt_length = key.byte_size - md.digest_size - 2
            salt_length = md.digest_size
        else:
            if key is None:
                salt_length = md.digest_size
            else:
                salt_length = padding.calculate_max_pss_salt_length(key, md)
        signer["signature_algorithm"] = algos.SignedDigestAlgorithm(
            {
                "algorithm": "rsassa_pss",
                "parameters": algos.RSASSAPSSParams(
                    {
                        "hash_algorithm": algos.DigestAlgorithm(
                            {"algorithm": hashalgo.lower()}
                        ),
                        "mask_gen_algorithm": algos.MaskGenAlgorithm(
                            {
                                "algorithm": algos.MaskGenAlgorithmId("mgf1"),
                                "parameters": {
                                    "algorithm": algos.DigestAlgorithmId(hashalgo.lower()),
                                },
                            }
                        ),
                        "salt_length": algos.Integer(salt_length),
                        "trailer_field": algos.TrailerField(1),
                    }
                ),
            }
        )

    if attrs:
        if attrs is True:
            signing_certificate1 = cms.CMSAttribute(
                {
                    "type": cms.CMSAttributeType("signing_certificate"),
                    "values": (
                        tsp.SigningCertificate(
                            {
                                "certs": [
                                    tsp.ESSCertID(
                                        {
                                            "cert_hash": hashlib.sha1(
                                                cert.dump()
                                            ).digest(),
                                            "issuer_serial": tsp.IssuerSerial(
                                                {
                                                    "issuer": (
                                                        x509.GeneralName(
                                                            {
                                                                "directory_name": cert.issuer,
                                                            }
                                                        ),
                                                    ),
                                                    "serial_number": cert.serial_number,
                                                }
                                            ),
                                        }
                                    ),
                                ]
                            }
                        ),
                    ),
                }
            )

            signing_certificate2 = cms.CMSAttribute(
                {
                    "type": cms.CMSAttributeType("signing_certificate_v2"),
                    "values": [
                        tsp.SigningCertificateV2(
                            {
                                "certs": [
                                    tsp.ESSCertIDv2(
                                        {
                                            "hash_algorithm": algos.DigestAlgorithm(
                                                {"algorithm": "sha256"}
                                            ),
                                            "cert_hash": hashlib.sha256(
                                                cert.dump()
                                            ).digest(),
                                            "issuer_serial": tsp.IssuerSerial(
                                                {
                                                    "issuer": (
                                                        x509.GeneralName(
                                                            {
                                                                "directory_name": cert.issuer,
                                                            }
                                                        ),
                                                    ),
                                                    "serial_number": cert.serial_number,
                                                }
                                            ),
                                        }
                                    ),
                                ]
                            }
                        ),
                    ],
                }
            )

            signer["signed_attrs"] = [
                cms.CMSAttribute(
                    {
                        "type": cms.CMSAttributeType("content_type"),
                        "values": ("data",),
                    }
                ),
                cms.CMSAttribute(
                    {
                        "type": cms.CMSAttributeType("message_digest"),
                        "values": (signed_value,),
                    }
                ),
                # cms.CMSAttribute(
                #    {
                #        "type": cms.CMSAttributeType("signing_time"),
                #        "values": (cms.Time({"utc_time": core.UTCTime(signed_time)}),),
                #    }
                # ),
                signing_certificate2,
            ]
        else:
            if isinstance(attrs, types.FunctionType):
                attrs = attrs(signed_value)
            signer["signed_attrs"] = attrs

    config = {
        "version": "v1",
        "digest_algorithms": cms.DigestAlgorithms(
            (algos.DigestAlgorithm({"algorithm": hashalgo}),)
        ),
        "encap_content_info": {
            "content_type": "data",
        },
        "certificates": certificates,
        "signer_infos": [
            signer,
        ],
    }
    if ocspurl and ocspissuer:
        from cryptography.hazmat.backends.openssl.backend import backend
        from cryptography.x509 import ocsp as cocsp
        from cryptography import x509 as cx509

        ocspuser = cert.dump()
        ocspuser = cx509.load_der_x509_certificate(ocspuser, backend=backend)

        builder = cocsp.OCSPRequestBuilder()
        builder = builder.add_certificate(ocspuser, ocspissuer, hashes.SHA1())
        req = builder.build()
        data = req.public_bytes(serialization.Encoding.DER)

        response = requests.post(
            ocspurl,
            headers={"Content-Type": "application/ocsp-request"},
            data=data,
        )
        data = ocsp.OCSPResponse.load(response.content)
        other = cms.RevocationInfoChoice(
            {
                "other": cms.OtherRevocationInfoFormat(
                    {
                        "other_rev_info_format": cms.OtherRevInfoFormatId(
                            "ocsp_response"
                        ),
                        "other_rev_info": data,
                    }
                )
            }
        )
        config["crls"] = cms.RevocationInfoChoices([other])

    datas = cms.ContentInfo(
        {
            "content_type": cms.ContentType("signed_data"),
            "content": cms.SignedData(config),
        }
    )
    if attrs:
        tosign = datas["content"]["signer_infos"][0]["signed_attrs"].dump()
        tosign = b"\x31" + tosign[1:]
    else:
        tosign = datau
    if hsm is not None:
        signed_value_signature = hsm.sign(keyid, tosign, hashalgo)
    elif isinstance(key, keys.PrivateKeyInfo):
        key = asymmetric.load_private_key(key)
        if pss:
            signed_value_signature = asymmetric.rsa_pss_sign(key, tosign, hashalgo.lower())
        else:
            signed_value_signature = asymmetric.rsa_pkcs1v15_sign(
                key, tosign, hashalgo.lower()
            )
    else:
        if pss:
            md = getattr(hashes, hashalgo.upper())
            hasher = hashes.Hash(md(), backend=backends.default_backend())
            hasher.update(tosign)
            digest = hasher.finalize()
            signed_value_signature = key.sign(
                digest,
                padding.PSS(mgf=padding.MGF1(md()), salt_length=salt_length),
                utils.Prehashed(md()),
            )
        else:
            if isinstance(key, ec.EllipticCurvePrivateKey):
                signed_value_signature = key.sign(
                    tosign, ec.ECDSA(getattr(hashes, hashalgo.upper())())
                )
            else:
                signed_value_signature = key.sign(
                    tosign, padding.PKCS1v15(), getattr(hashes, hashalgo.upper())()
                )

    if timestampurl is not None:
        datas["content"]["signer_infos"][0]["unsigned_attrs"] = timestamp(
            signed_value_signature,
            hashalgo,
            timestampurl,
            timestampcredentials,
            timestamp_req_options,
        )

    # signed_value_signature = core.OctetString(signed_value_signature)
    datas["content"]["signer_infos"][0]["signature"] = signed_value_signature

    # open('signed-content-info', 'wb').write(datas.dump())
    return datas.dump()



#!/usr/bin/env vpython3
from pdf_annotate import PdfAnnotator, Location, Appearance

annotationtext = "some text"

a = PdfAnnotator("pdf.pdf")
a.add_annotation(
    "text",
    Location(x1=50, y1=50, x2=200, y2=100, page=0),
    Appearance(
        fill=(0, 0, 0),
        stroke_width=1,
        wrap_text=True,
        font_size=12,
        content=annotationtext,
    ),
)
a.write("pdf-a.pdf")



#!/usr/bin/env vpython3
import sys
import pdfbox
from jnius import autoclass


pdfname = 'pdf.pdf'
if len(sys.argv) > 1:
    pdfname = sys.argv[1]
parser = autoclass('org.apache.pdfbox.preflight.parser.PreflightParser')(pdfname)
parser.parse()
#org.apache.pdfbox.preflight.PreflightDocument
document = parser.getPreflightDocument()
document.validate()
result = document.getResult()
document.close()
if result.isValid():
    print("The file:{} is a valid PDF/A-1b file".format(pdfname))
else:
    print("The file: {} is not valid, error(s) :".format(pdfname))
    for error in result.getErrorsList():
        print(error.getErrorCode(), " : ", error.getDetails())



#!/usr/bin/env vpython3
# coding: utf-8

import os
import sys
import binascii
import datetime
import PyKCS11
import base64
import hashlib

from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization

from asn1crypto import x509 as asn1x509
from asn1crypto import keys as asn1keys
from asn1crypto import pem as asn1pem
from asn1crypto import util as asn1util

import paramiko.agent

import cryptography


class BaseHSM:
    def certificate(self):
        """
        callback for HSM
        used to identfy the ssh agents key exports via fingerprint

        :return: public-key-fingerprint, certificate-in-pem
        """
        raise NotImplementedError()

    def sign(self, keyid, data, mech):
        """
        sign

        :param keyid: the keyid as returned by certificate()
        :param data:
        :param mech: hash algo
        :return: PKCS7 signature blob
        """
        raise NotImplementedError()


class HSM(BaseHSM):
    def __init__(self, dllpath):
        self.pkcs11 = PyKCS11.PyKCS11Lib()
        self.pkcs11.load(dllpath)
        self.session = None

    def getSlot(self, label):
        slots = self.pkcs11.getSlotList(tokenPresent=True)
        for slot in slots:
            info = self.pkcs11.getTokenInfo(slot)
            try:
                if info.label.split("\0")[0].strip() == label:
                    return slot
            except AttributeError:
                continue
        return None

    def create(self, label, pin, sopin):
        slot = self.getSlot(label)
        if slot is not None:
            return
        slot = self.pkcs11.getSlotList(tokenPresent=True)[-1]
        self.pkcs11.initToken(slot, sopin, label)
        session = self.pkcs11.openSession(
            slot, PyKCS11.CKF_SERIAL_SESSION | PyKCS11.CKF_RW_SESSION
        )
        session.login(sopin, user_type=PyKCS11.CKU_SO)
        session.initPin(pin)
        session.logout()
        session.closeSession()

    def login(self, label, pin):
        slot = self.getSlot(label)
        if slot is None:
            return
        self.session = self.pkcs11.openSession(
            slot, PyKCS11.CKF_SERIAL_SESSION | PyKCS11.CKF_RW_SESSION
        )
        self.session.login(pin)

    def logout(self):
        if self.session is not None:
            self.session.logout()
            self.session.closeSession()
            self.session = None

    def gen_privkey(self, label, key_id, key_length=2048):
        # label - just a label for identifying objects
        # key_id has to be the same for both objects, it will also be necessary
        #     when importing the certificate, to ensure it is linked with these keys.
        # key_length - key-length in bits

        public_template = [
            (PyKCS11.CKA_CLASS, PyKCS11.CKO_PUBLIC_KEY),
            (PyKCS11.CKA_TOKEN, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_PRIVATE, PyKCS11.CK_FALSE),
            (PyKCS11.CKA_MODULUS_BITS, key_length),
            #            (PyKCS11.CKA_PUBLIC_EXPONENT, (0x01, 0x00, 0x01)),
            (PyKCS11.CKA_ENCRYPT, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_VERIFY, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_VERIFY_RECOVER, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_WRAP, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_LABEL, label),
            (PyKCS11.CKA_ID, key_id)
            #            (PyKCS11.CKA_KEY_TYPE, PyKCS11.CKK_RSA),
            #            (PyKCS11.CKA_SENSITIVE, PyKCS11.CK_FALSE),
        ]

        private_template = [
            (PyKCS11.CKA_CLASS, PyKCS11.CKO_PRIVATE_KEY),
            (PyKCS11.CKA_TOKEN, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_PRIVATE, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_DECRYPT, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_SIGN, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_SIGN_RECOVER, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_UNWRAP, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_LABEL, label),
            (PyKCS11.CKA_ID, key_id)
            #            (PyKCS11.CKA_SENSITIVE, PyKCS11.CK_TRUE),
        ]

        self.session.generateKeyPair(public_template, private_template)

    def cert_save(self, cert, label, subject, key_id):
        cert_template = [
            (PyKCS11.CKA_CLASS, PyKCS11.CKO_CERTIFICATE),
            (PyKCS11.CKA_CERTIFICATE_TYPE, PyKCS11.CKC_X_509),
            (PyKCS11.CKA_TOKEN, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_LABEL, label.encode("utf-8")),
            (
                PyKCS11.CKA_ID,
                key_id,
            ),  # must be set, and DER see Table 24, X.509 Certificate Object Attributes
            (
                PyKCS11.CKA_SUBJECT,
                subject.encode("utf-8"),
            ),  # must be set and DER, see Table 24, X.509 Certificate Object Attributes
            # (PyKCS11.CKA_PRIVATE, PyKCS11.CK_FALSE),
            # (PyKCS11.CKA_TRUSTED, PyKCS11.CK_TRUE),
            # (PyKCS11.CKA_SENSITIVE, PyKCS11.CK_FALSE),
            # (PyKCS11.CKA_ENCRYPT, PyKCS11.CK_TRUE),
            # (PyKCS11.CKA_VERIFY, PyKCS11.CK_TRUE),
            # (PyKCS11.CKA_MODIFIABLE, PyKCS11.CK_TRUE),
            #            (PyKCS11.CKA_ISSUER, cert.Issuer);
            #            (PyKCS11.CKA_SERIAL_NUMBER,cert.SerialNumber)
            (PyKCS11.CKA_VALUE, cert),  # must be BER-encoded
        ]

        self.session.createObject(cert_template)

    def cert_load(self, keyID):
        rec = self.session.findObjects(
            [(PyKCS11.CKA_CLASS, PyKCS11.CKO_CERTIFICATE), (PyKCS11.CKA_ID, keyID)]
        )
        if len(rec) == 0:
            return None
        value = bytes(rec[0].to_dict()["CKA_VALUE"])
        return value

    def certsign(self, sn, pubKey, subject, until, caprivKey, ca):
        args = {
            "version": "v1",
            "serial_number": sn,
            "issuer": asn1x509.Name.build(
                {
                    "common_name": "hsm CA",
                }
            ),
            "subject": asn1x509.Name.build(
                {
                    "common_name": subject,
                }
            ),
            "signature": {
                "algorithm": "sha256_rsa",
                "parameters": None,
            },
            "validity": {
                "not_before": asn1x509.Time(
                    {
                        "general_time": datetime.datetime.now(tz=asn1util.timezone.utc)
                        - datetime.timedelta(days=1),
                    }
                ),
                "not_after": asn1x509.Time(
                    {
                        "general_time": until,
                    }
                ),
            },
            "subject_public_key_info": {
                "algorithm": {
                    "algorithm": "rsa",
                    "parameters": None,
                },
                "public_key": pubKey,
            },
        }
        if ca:
            args.update(
                {
                    "extensions": [
                        {
                            "extn_id": "basic_constraints",
                            "critical": True,
                            "extn_value": {"ca": True, "path_len_constraint": None},
                        },
                        {
                            "extn_id": "key_usage",
                            "critical": True,
                            "extn_value": set(
                                [
                                    "crl_sign",
                                    "digital_signature",
                                    "key_cert_sign",
                                ]
                            ),
                        },
                    ]
                }
            )
        else:
            args.update(
                {
                    "extensions": [
                        {
                            "extn_id": "basic_constraints",
                            "critical": True,
                            "extn_value": {"ca": False},
                        },
                        {
                            "extn_id": "key_usage",
                            "critical": True,
                            "extn_value": set(
                                [
                                    "digital_signature",
                                    "key_agreement",
                                    "key_encipherment",
                                    "non_repudiation",
                                ]
                            ),
                        },
                    ]
                }
            )
        tbs = asn1x509.TbsCertificate(args)

        # Sign the TBS Certificate
        data = tbs.dump()
        value = self.session.sign(
            caprivKey, data, PyKCS11.Mechanism(PyKCS11.CKM_SHA256_RSA_PKCS, None)
        )
        value = bytes(bytearray(value))

        cert = asn1x509.Certificate(
            {
                "tbs_certificate": tbs,
                "signature_algorithm": {
                    "algorithm": "sha256_rsa",
                    "parameters": None,
                },
                "signature_value": value,
            }
        )
        return cert.dump()

    def ca_gen(self, label, keyID, subject):
        privKey = self.session.findObjects(
            [(PyKCS11.CKA_CLASS, PyKCS11.CKO_PRIVATE_KEY), (PyKCS11.CKA_ID, keyID)]
        )[0]
        pubKey = self.session.findObjects(
            [(PyKCS11.CKA_CLASS, PyKCS11.CKO_PUBLIC_KEY), (PyKCS11.CKA_ID, keyID)]
        )[0]

        modulus = self.session.getAttributeValue(pubKey, [PyKCS11.CKA_MODULUS])[0]
        modulus = binascii.hexlify(bytearray(modulus)).decode("utf-8")
        exponent = self.session.getAttributeValue(
            pubKey, [PyKCS11.CKA_PUBLIC_EXPONENT]
        )[0]
        exponent = binascii.hexlify(bytearray(exponent)).decode("utf-8")
        pubKey = asn1keys.RSAPublicKey(
            {
                "modulus": int("0x" + modulus, 16),
                "public_exponent": int("0x" + exponent, 16),
            }
        )
        # pubKey = asn1keys.RSAPublicKey.load(pubKey.dump())
        until = datetime.datetime.now(tz=asn1util.timezone.utc) + datetime.timedelta(
            days=365 * 10
        )
        der_bytes = self.certsign(1, pubKey, subject, until, privKey, True)
        self.cert_save(der_bytes, label, subject, keyID)

    def ca_sign(self, keyID, label, sn, subject, days, cakeyID):
        caprivKey = self.session.findObjects(
            [(PyKCS11.CKA_CLASS, PyKCS11.CKO_PRIVATE_KEY), (PyKCS11.CKA_ID, cakeyID)]
        )[0]

        pubKey = self.session.findObjects(
            [(PyKCS11.CKA_CLASS, PyKCS11.CKO_PUBLIC_KEY), (PyKCS11.CKA_ID, keyID)]
        )[0]
        modulus = self.session.getAttributeValue(pubKey, [PyKCS11.CKA_MODULUS])[0]
        modulus = binascii.hexlify(bytearray(modulus)).decode("utf-8")
        exponent = self.session.getAttributeValue(
            pubKey, [PyKCS11.CKA_PUBLIC_EXPONENT]
        )[0]
        exponent = binascii.hexlify(bytearray(exponent)).decode("utf-8")
        pubKey = asn1keys.RSAPublicKey(
            {
                "modulus": int("0x" + modulus, 16),
                "public_exponent": int("0x" + exponent, 16),
            }
        )
        # pubKey = asn1keys.RSAPublicKey.load(pubKey.dump())
        until = datetime.datetime.now(tz=asn1util.timezone.utc) + datetime.timedelta(
            days=days
        )
        der_bytes = self.certsign(sn, pubKey, subject, until, caprivKey, False)
        self.cert_save(der_bytes, label, subject, keyID)

    def cert_export(self, fname, keyID):
        der_bytes = self.cert_load(keyID)
        pem_bytes = asn1pem.armor("CERTIFICATE", der_bytes)
        with open(fname + ".der", "wb") as fp:
            fp.write(der_bytes)
        with open(fname + ".pem", "wb") as fp:
            fp.write(pem_bytes)


class SSHAgentHSM(BaseHSM):
    def __init__(self, cert):
        assert isinstance(cert, cryptography.x509.Certificate)
        self._a = paramiko.agent.Agent()
        self._cert = cert

    def close(self):
        self._a.close()

    def certificate(self):
        """
        callback for HSM
        used to identfy the ssh agents key exports via fingerprint

        :return: public-key-fingerprint, certificate-in-pem
        """

        # https://superuser.com/questions/421997/what-is-a-ssh-key-fingerprint-and-how-is-it-generated
        # convert RSA Key to SSH Fingerprint
        alg, key = (
            self._cert.public_key()
            .public_bytes(
                encoding=cryptography.hazmat.primitives.serialization.Encoding.OpenSSH,
                format=cryptography.hazmat.primitives.serialization.PublicFormat.OpenSSH,
            )
            .split(b" ")
        )

        fp = b"SHA256:" + base64.b64encode(
            hashlib.sha256(base64.b64decode(key)).digest()
        )
        cert = self._cert.public_bytes(
            cryptography.hazmat.primitives.serialization.Encoding.PEM
        )

        return fp, cert

    @staticmethod
    def _decode_fp(keyfp):
        """
        decode a fingerprint

        :param keyfp: key fingerprint in OpenSSH Format
        :return: alg, fingerprint-binary
        """
        if not isinstance(keyfp, str):
            keyfp = keyfp.decode()
        alg, other = keyfp.split(":", 1)
        if alg == "SHA256":
            # pad base64 data
            data = other.encode() + b"=" * (-len(other) % 4)
            fp = base64.b64decode(data)
        elif alg == "MD5":
            data = other.replace(":", " ")
            fp = bytes.fromhex(data)
        else:
            raise ValueError(alg)
        return alg.lower(), fp

    def key(self, fp):
        """
        lookup a ssh-agent-exported key using fingerprint

        :param fp: the fingerprint
        :return: the key on success
        """

        alg, fp = self._decode_fp(fp)
        for key in self._a.get_keys():
            kfp = getattr(hashlib, alg)(key.asbytes()).digest()
            if kfp == fp:
                break
        else:
            raise ValueError("Key not found")
        return key

    def sign(self, keyid, data, hashalgo):
        """
        sign using ssh-agent sign_data
        creates RSA signature with padding=PKCS1v15 alg=SHA1

        :param keyid: the keyid as returned by certificate()
        :param data:
        :param hashalgo: has to be sha1, sha256 or sha512
        :return: PKCS7 signature blob
        """
        assert hashalgo in ("sha1", "sha256", "sha512")

        if not isinstance(data, bytes):
            data = data.encode()

        # defined in
        # SSH Agent Protocol draft-miller-ssh-agent-00 5.3.  Signature flags
        # https://tools.ietf.org/html/draft-miller-ssh-agent-00#section-5.3
        flags = {
            "sha1": 0,
            "sha256": 2,  # SSH_AGENT_RSA_SHA2_256
            "sha512": 4,  # SSH_AGENT_RSA_SHA2_512
        }[hashalgo]

        key = self.key(keyid)

        # AgentKey.sign_ssh_data is padding=PKCS1v15 alg=SHA1
        # paramiko does not expose the ssh-agent sign flags to use sha2-256/512
        # re-implement sign_ssh_agent ..
        msg = paramiko.message.Message()
        msg.add_byte(paramiko.agent.cSSH2_AGENTC_SIGN_REQUEST)
        msg.add_string(key.blob)
        msg.add_string(data)
        msg.add_int(flags)
        ptype, result = self._a._send_message(msg)
        if ptype != paramiko.agent.SSH2_AGENT_SIGN_RESPONSE:
            raise paramiko.SSHException("key cannot be used for signing")
        d = paramiko.message.Message(result.get_binary())

        # parse operation result
        alg = d.get_text()

        # interpret
        if alg in ("ssh-rsa", "rsa-sha2-256", "rsa-sha2-512"):
            sig = d.get_binary()
        else:
            raise ValueError(alg)
        return sig



#******************************************************************************
# TTFontFile class                                                             
#                                                                              
# This class is based on The ReportLab Open Source PDF library                 
# written in Python - http://www.reportlab.com/software/opensource/            
# together with ideas from the OpenOffice source code and others.              
#                                                                              
# Version:  1.04                                                               
# Date:     2011-09-18                                                         
# Author:   Ian Back <ianb@bpm1.com>                                           
# License:  LGPL                                                               
# Copyright (c) Ian Back, 2010                                                 
# Ported to Python 2.7 by Mariano Reingart (reingart@gmail.com) on 2012        
# This header must be retained in any redistribution or                        
# modification of the file.                                                    
#                                                                              
#******************************************************************************

from struct import pack, unpack, unpack_from
import re
import warnings
from .php import die, substr, str_repeat, str_pad, strlen, count
from .py3k import b, ord


# Define the value used in the "head" table of a created TTF file
# 0x74727565 "true" for Mac
# 0x00010000 for Windows
# Either seems to work for a font embedded in a PDF file
# when read by Adobe Reader on a Windows PC(!)
_TTF_MAC_HEADER = False


# TrueType Font Glyph operators
GF_WORDS = (1 << 0)
GF_SCALE = (1 << 3)
GF_MORE  = (1 << 5)
GF_XYSCALE  = (1 << 6)
GF_TWOBYTWO = (1 << 7)


def sub32(x, y):
    xlo = x[1]
    xhi = x[0]
    ylo = y[1]
    yhi = y[0]
    if (ylo > xlo):  
        xlo += 1 << 16 
        yhi += 1 
    reslo = xlo-ylo
    if (yhi > xhi):  
        xhi += 1 << 16  
    reshi = xhi-yhi
    reshi = reshi & 0xFFFF
    return (reshi, reslo)

def calcChecksum(data): 
    if (strlen(data) % 4):
        data += str_repeat(b("\0"), (4-(len(data) % 4)))
    hi=0x0000
    lo=0x0000
    for i in range(0, len(data), 4): 
        hi += (ord(data[i])<<8) + ord(data[i+1])
        lo += (ord(data[i+2])<<8) + ord(data[i+3])
        hi += lo >> 16
        lo = lo & 0xFFFF
        hi = hi & 0xFFFF
    return (hi, lo)


class TTFontFile:

    def __init__(self):
        self.maxStrLenRead = 200000    # Maximum size of glyf table to read in as string (otherwise reads each glyph from file)

    def getMetrics(self, file):
        self.filename = file
        self.fh = open(file,'rb')
        self._pos = 0
        self.charWidths = []
        self.glyphPos = {}
        self.charToGlyph = {}
        self.tables = {}
        self.otables = {}
        self.ascent = 0
        self.descent = 0
        self.TTCFonts = {}
        self.version = version = self.read_ulong()
        if (version==0x4F54544F):
            die("Postscript outlines are not supported")
        if (version==0x74746366):
            die("ERROR - TrueType Fonts Collections not supported")
        if (version not in (0x00010000,0x74727565)):
            die("Not a TrueType font: version=" + version)
        self.readTableDirectory()
        self.extractInfo()
        self.fh.close()
    
    def readTableDirectory(self, ):
        self.numTables = self.read_ushort()
        self.searchRange = self.read_ushort()
        self.entrySelector = self.read_ushort()
        self.rangeShift = self.read_ushort()
        self.tables = {}    
        for i in range(self.numTables):
            record = {}
            record['tag'] = self.read_tag()
            record['checksum'] = (self.read_ushort(),self.read_ushort())
            record['offset'] = self.read_ulong()
            record['length'] = self.read_ulong()
            self.tables[record['tag']] = record    

    def get_table_pos(self, tag):
        offset = self.tables[tag]['offset']
        length = self.tables[tag]['length']
        return (offset, length)
    
    def seek(self, pos): 
        self._pos = pos
        self.fh.seek(self._pos)
    
    def skip(self, delta): 
        self._pos = self._pos + delta
        self.fh.seek(self._pos)
    
    def seek_table(self, tag, offset_in_table = 0):
        tpos = self.get_table_pos(tag)
        self._pos = tpos[0] + offset_in_table
        self.fh.seek(self._pos)
        return self._pos

    def read_tag(self):
        self._pos += 4
        return self.fh.read(4).decode("latin1")

    def read_short(self): 
        self._pos += 2
        s = self.fh.read(2)
        a = (ord(s[0])<<8) + ord(s[1])
        if (a & (1 << 15) ):
            a = (a - (1 << 16)) 
        return a
    
    def unpack_short(self, s):
        a = (ord(s[0])<<8) + ord(s[1])
        if (a & (1 << 15) ):
            a = (a - (1 << 16))     
        return a
    
    def read_ushort(self):
        self._pos += 2
        s = self.fh.read(2)
        return (ord(s[0])<<8) + ord(s[1])

    def read_ulong(self): 
        self._pos += 4
        s = self.fh.read(4)
        # if large uInt32 as an integer, PHP converts it to -ve
        return (ord(s[0])*16777216) + (ord(s[1])<<16) + (ord(s[2])<<8) + ord(s[3]) #     16777216  = 1<<24

    def get_ushort(self, pos): 
        self.fh.seek(pos)
        s = self.fh.read(2)
        return (ord(s[0])<<8) + ord(s[1])

    def get_ulong(self, pos):
        self.fh.seek(pos)
        s = self.fh.read(4)
        # iF large uInt32 as an integer, PHP converts it to -ve
        return (ord(s[0])*16777216) + (ord(s[1])<<16) + (ord(s[2])<<8) + ord(s[3]) #     16777216  = 1<<24    

    def pack_short(self, val):
        if (val<0):
            val = abs(val)
            val = ~val
            val += 1
        return pack(">H",val) 
    
    def splice(self, stream, offset, value):
        return substr(stream,0,offset) + value + substr(stream,offset+strlen(value))
    
    def _set_ushort(self, stream, offset, value):
        up = pack(">H", value)
        return self.splice(stream, offset, up)    

    def _set_short(self, stream, offset, val):
        if (val<0):
            val = abs(val)
            val = ~val
            val += 1
        up = pack(">H",val) 
        return self.splice(stream, offset, up)

    def get_chunk(self, pos, length): 
        self.fh.seek(pos)
        if (length <1):  return '' 
        return (self.fh.read(length))

    def get_table(self, tag):
        (pos, length) = self.get_table_pos(tag)
        if (length == 0):
            die('Truetype font (' + self.filename + '): error reading table: ' + tag) 
        self.fh.seek(pos)
        return (self.fh.read(length))

    def add(self, tag, data):
        if (tag == 'head') :
            data = self.splice(data, 8, b("\0\0\0\0"))        
        self.otables[tag] = data

############################################/
############################################/

############################################/

    def extractInfo(self): 
        #################/
        # name - Naming table
        #################/
        self.sFamilyClass = 0
        self.sFamilySubClass = 0

        name_offset = self.seek_table("name")
        format = self.read_ushort()
        if (format != 0):
            die("Unknown name table format " + format)
        numRecords = self.read_ushort()
        string_data_offset = name_offset + self.read_ushort()
        names = {1:'',2:'',3:'',4:'',6:''}
        K = list(names.keys())
        nameCount = len(names)
        for i in range(numRecords): 
            platformId = self.read_ushort()
            encodingId = self.read_ushort()
            languageId = self.read_ushort()
            nameId = self.read_ushort()
            length = self.read_ushort()
            offset = self.read_ushort()
            if (nameId not in K): continue
            N = ''
            if (platformId == 3 and encodingId == 1 and languageId == 0x409):  # Microsoft, Unicode, US English, PS Name
                opos = self._pos
                self.seek(string_data_offset + offset)
                if (length % 2 != 0):
                    die("PostScript name is UTF-16BE string of odd length")
                length /= 2
                N = ''
                while (length > 0):
                    char = self.read_ushort()
                    N += (chr(char))
                    length -= 1
                self._pos = opos
                self.seek(opos)
            
            elif (platformId == 1 and encodingId == 0 and languageId == 0):  # Macintosh, Roman, English, PS Name
                opos = self._pos
                N = self.get_chunk(string_data_offset + offset, length).decode("latin1")
                self._pos = opos
                self.seek(opos)
            
            if (N and names[nameId]==''):
                names[nameId] = N
                nameCount -= 1
                if (nameCount==0): break
            
        
        if (names[6]):
            psName = names[6]
        elif (names[4]):
            psName = re.sub(' ','-',names[4])
        elif (names[1]):
            psName = re.sub(' ','-',names[1])
        else:
            psName = ''
        if (not psName):
            die("Could not find PostScript font name")
        self.name = psName
        if (names[1]):
            self.familyName = names[1]  
        else:  
            self.familyName = psName 
        if (names[2]):
            self.styleName = names[2]
        else:
            self.styleName = 'Regular' 
        if (names[4]):
            self.fullName = names[4]
        else:
            self.fullName = psName 
        if (names[3]):
            self.uniqueFontID = names[3]
        else:
            self.uniqueFontID = psName 
        if (names[6]):
            self.fullName = names[6] 

        #################/
        # head - Font header table
        #################/
        self.seek_table("head")
        self.skip(18) 
        self.unitsPerEm = unitsPerEm = self.read_ushort()
        scale = 1000 / float(unitsPerEm)
        self.skip(16)
        xMin = self.read_short()
        yMin = self.read_short()
        xMax = self.read_short()
        yMax = self.read_short()
        self.bbox = [(xMin*scale), (yMin*scale), (xMax*scale), (yMax*scale)]
        self.skip(3*2)
        indexToLocFormat = self.read_ushort()
        glyphDataFormat = self.read_ushort()
        if (glyphDataFormat != 0):
            die('Unknown glyph data format ' + glyphDataFormat)

        #################/
        # hhea metrics table
        #################/
        # ttf2t1 seems to use this value rather than the one in OS/2 - so put in for compatibility
        if ("hhea" in self.tables):
            self.seek_table("hhea")
            self.skip(4)
            hheaAscender = self.read_short()
            hheaDescender = self.read_short()
            self.ascent = (hheaAscender *scale)
            self.descent = (hheaDescender *scale)
        

        #################/
        # OS/2 - OS/2 and Windows metrics table
        #################/
        if ("OS/2" in self.tables): 
            self.seek_table("OS/2")
            version = self.read_ushort()
            self.skip(2)
            usWeightClass = self.read_ushort()
            self.skip(2)
            fsType = self.read_ushort()
            if (fsType == 0x0002 or (fsType & 0x0300) != 0): 
                die('ERROR - Font file ' + self.filename + ' cannot be embedded due to copyright restrictions.')
                self.restrictedUse = True
            
            self.skip(20)
            sF = self.read_short()
            self.sFamilyClass = (sF >> 8)
            self.sFamilySubClass = (sF & 0xFF)
            self._pos += 10  #PANOSE = 10 byte length
            panose = self.fh.read(10)
            self.skip(26)
            sTypoAscender = self.read_short()
            sTypoDescender = self.read_short()
            if (not self.ascent): 
                self.ascent = (sTypoAscender*scale)
            if (not self.descent): 
                self.descent = (sTypoDescender*scale)
            if (version > 1):
                self.skip(16)
                sCapHeight = self.read_short()
                self.capHeight = (sCapHeight*scale)
            else:
                self.capHeight = self.ascent            
        
        else:
            usWeightClass = 500
            if (not self.ascent): self.ascent = (yMax*scale)
            if (not self.descent): self.descent = (yMin*scale)
            self.capHeight = self.ascent
        
        self.stemV = 50 + int(pow((usWeightClass / 65.0),2))

        #################/
        # post - PostScript table
        #################/
        self.seek_table("post")
        self.skip(4) 
        self.italicAngle = self.read_short() + self.read_ushort() / 65536.0
        self.underlinePosition = self.read_short() * scale
        self.underlineThickness = self.read_short() * scale
        isFixedPitch = self.read_ulong()

        self.flags = 4

        if (self.italicAngle!= 0):
            self.flags = self.flags | 64
        if (usWeightClass >= 600):
            self.flags = self.flags | 262144
        if (isFixedPitch):
            self.flags = self.flags | 1

        #################/
        # hhea - Horizontal header table
        #################/
        self.seek_table("hhea")
        self.skip(32) 
        metricDataFormat = self.read_ushort()
        if (metricDataFormat != 0):
            die('Unknown horizontal metric data format '.metricDataFormat)
        numberOfHMetrics = self.read_ushort()
        if (numberOfHMetrics == 0):
            die('Number of horizontal metrics is 0')

        #################/
        # maxp - Maximum profile table
        #################/
        self.seek_table("maxp")
        self.skip(4)
        numGlyphs = self.read_ushort()

        #################/
        # cmap - Character to glyph index mapping table
        #################/
        cmap_offset = self.seek_table("cmap")
        self.skip(2)
        cmapTableCount = self.read_ushort()
        unicode_cmap_offset = 0
        unicode_cmap_offset12 = 0
        
        for i in range(cmapTableCount):
            platformID = self.read_ushort()
            encodingID = self.read_ushort()
            offset = self.read_ulong()
            save_pos = self._pos
            if platformID == 3 and encodingID == 10:  # Microsoft, UCS-4
                format = self.get_ushort(cmap_offset + offset)
                if (format == 12):
                    if not unicode_cmap_offset12:
                        unicode_cmap_offset12 = cmap_offset + offset
                    break
            if ((platformID == 3 and encodingID == 1) or platformID == 0):  # Microsoft, Unicode
                format = self.get_ushort(cmap_offset + offset)
                if (format == 4):
                    if (not unicode_cmap_offset):
                        unicode_cmap_offset = cmap_offset + offset
                    break
                    
            self.seek(save_pos)
        
        if not unicode_cmap_offset and not unicode_cmap_offset12:
            die('Font (' + self.filename + ') does not have cmap for Unicode (platform 3, encoding 1, format 4, or platform 3, encoding 10, format 12, or platform 0, any encoding, format 4)')

        glyphToChar = {}
        charToGlyph = {}
        if unicode_cmap_offset12:
            self.getCMAP12(unicode_cmap_offset12, glyphToChar, charToGlyph)
        else:    
            self.getCMAP4(unicode_cmap_offset, glyphToChar, charToGlyph)

        #################/
        # hmtx - Horizontal metrics table
        #################/
        self.getHMTX(numberOfHMetrics, numGlyphs, glyphToChar, scale)


############################################/
############################################/

    def makeSubset(self, file, subset):
        self.filename = file
        self.fh = open(file ,'rb')
        self._pos = 0
        self.charWidths = []
        self.glyphPos = {}
        self.charToGlyph = {}
        self.tables = {}
        self.otables = {}
        self.ascent = 0
        self.descent = 0
        self.skip(4)
        self.maxUni = 0
        self.readTableDirectory()

        #################/
        # head - Font header table
        #################/
        self.seek_table("head")
        self.skip(50) 
        indexToLocFormat = self.read_ushort()
        glyphDataFormat = self.read_ushort()

        #################/
        # hhea - Horizontal header table
        #################/
        self.seek_table("hhea")
        self.skip(32) 
        metricDataFormat = self.read_ushort()
        orignHmetrics = numberOfHMetrics = self.read_ushort()

        #################/
        # maxp - Maximum profile table
        #################/
        self.seek_table("maxp")
        self.skip(4)
        numGlyphs = self.read_ushort()

        #################/
        # cmap - Character to glyph index mapping table
        #################/
        cmap_offset = self.seek_table("cmap")
        self.skip(2)
        cmapTableCount = self.read_ushort()
        unicode_cmap_offset = 0
        unicode_cmap_offset12 = 0
        for i in range(cmapTableCount):
            platformID = self.read_ushort()
            encodingID = self.read_ushort()
            offset = self.read_ulong()
            save_pos = self._pos
            if platformID == 3 and encodingID == 10:  # Microsoft, UCS-4
                format = self.get_ushort(cmap_offset + offset)
                if (format == 12):
                    if not unicode_cmap_offset12:
                        unicode_cmap_offset12 = cmap_offset + offset
                    break
            if ((platformID == 3 and encodingID == 1) or platformID == 0):  # Microsoft, Unicode
                format = self.get_ushort(cmap_offset + offset)
                if (format == 4):
                    unicode_cmap_offset = cmap_offset + offset
                    break
                
            self.seek(save_pos )
        
        if not unicode_cmap_offset and not unicode_cmap_offset12:
            die('Font (' + self.filename + ') does not have cmap for Unicode (platform 3, encoding 1, format 4, or platform 3, encoding 10, format 12, or platform 0, any encoding, format 4)')

        glyphToChar = {}
        charToGlyph = {}
        if unicode_cmap_offset12:
            self.getCMAP12(unicode_cmap_offset12, glyphToChar, charToGlyph)
        else:    
            self.getCMAP4(unicode_cmap_offset, glyphToChar, charToGlyph)

        self.charToGlyph = charToGlyph

        #################/
        # hmtx - Horizontal metrics table
        #################/
        scale = 1    # not used
        self.getHMTX(numberOfHMetrics, numGlyphs, glyphToChar, scale)

        #################/
        # loca - Index to location
        #################/
        self.getLOCA(indexToLocFormat, numGlyphs)

        subsetglyphs = [(0, 0)]     # special "sorted dict"!
        subsetCharToGlyph = {}
        for code in subset: 
            if (code in self.charToGlyph):
                if (self.charToGlyph[code], code) not in subsetglyphs:
                    subsetglyphs.append((self.charToGlyph[code], code))   # Old Glyph ID => Unicode
                subsetCharToGlyph[code] = self.charToGlyph[code]    # Unicode to old GlyphID
            self.maxUni = max(self.maxUni, code)
        (start,dummy) = self.get_table_pos('glyf')

        subsetglyphs.sort()
        glyphSet = {}
        n = 0
        fsLastCharIndex = 0    # maximum Unicode index (character code) in this font, according to the cmap subtable for platform ID 3 and platform- specific encoding ID 0 or 1.
        for originalGlyphIdx, uni in subsetglyphs:
            fsLastCharIndex = max(fsLastCharIndex , uni)
            glyphSet[originalGlyphIdx] = n    # old glyphID to new glyphID
            n += 1

        codeToGlyph = {}
        for uni, originalGlyphIdx in sorted(subsetCharToGlyph.items()):
            codeToGlyph[uni] = glyphSet[originalGlyphIdx] 
        
        self.codeToGlyph = codeToGlyph
        
        for originalGlyphIdx, uni in subsetglyphs: 
            nonlocals = {'start': start, 'glyphSet': glyphSet, 
                         'subsetglyphs': subsetglyphs}
            self.getGlyphs(originalGlyphIdx, nonlocals)

        numGlyphs = numberOfHMetrics = len(subsetglyphs)

        #tables copied from the original
        tags = ['name']
        for tag in tags:  
            self.add(tag, self.get_table(tag)) 
        tags = ['cvt ', 'fpgm', 'prep', 'gasp']
        for tag in tags:
            if (tag in self.tables):  
                self.add(tag, self.get_table(tag))        

        # post - PostScript
        opost = self.get_table('post')
        post = b("\x00\x03\x00\x00") + substr(opost,4,12) + b("\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00")
        self.add('post', post)

        # Sort CID2GID map into segments of contiguous codes
        if 0 in codeToGlyph:
            del codeToGlyph[0]
        #unset(codeToGlyph[65535])
        rangeid = 0
        range_ = {}
        prevcid = -2
        prevglidx = -1
        # for each character
        for cid, glidx in sorted(codeToGlyph.items()):
            if (cid == (prevcid + 1) and glidx == (prevglidx + 1)):
                range_[rangeid].append(glidx)
            else:
                # new range
                rangeid = cid
                range_[rangeid] = []
                range_[rangeid].append(glidx)
            prevcid = cid
            prevglidx = glidx

        # cmap - Character to glyph mapping - Format 4 (MS / )
        segCount = len(range_) + 1    # + 1 Last segment has missing character 0xFFFF
        searchRange = 1
        entrySelector = 0
        while (searchRange * 2 <= segCount ):
            searchRange = searchRange * 2
            entrySelector = entrySelector + 1
        
        searchRange = searchRange * 2
        rangeShift = segCount * 2 - searchRange
        length = 16 + (8*segCount ) + (numGlyphs+1)
        cmap = [0, 1,        # Index : version, number of encoding subtables
            3, 1,                # Encoding Subtable : platform (MS=3), encoding (Unicode)
            0, 12,            # Encoding Subtable : offset (hi,lo)
            4, length, 0,         # Format 4 Mapping subtable: format, length, language
            segCount*2,
            searchRange,
            entrySelector,
            rangeShift]

        range_ = sorted(range_.items())
        
        # endCode(s)
        for start, subrange in range_:
            endCode = start + (len(subrange)-1)
            cmap.append(endCode)    # endCode(s)
        
        cmap.append(0xFFFF)    # endCode of last Segment
        cmap.append(0)    # reservedPad

        # startCode(s)
        for start, subrange in range_: 
            cmap.append(start)    # startCode(s)
        
        cmap.append(0xFFFF)    # startCode of last Segment
        # idDelta(s) 
        for start, subrange in range_: 
            idDelta = -(start-subrange[0])
            n += count(subrange)
            cmap.append(idDelta)    # idDelta(s)
        
        cmap.append(1)    # idDelta of last Segment
        # idRangeOffset(s) 
        for subrange in range_: 
            cmap.append(0)    # idRangeOffset[segCount]      Offset in bytes to glyph indexArray, or 0
        
        cmap.append(0)    # idRangeOffset of last Segment
        for subrange, glidx in range_: 
            cmap.extend(glidx)
        
        cmap.append(0)    # Mapping for last character
        cmapstr = b('')
        for cm in cmap:
            if cm >= 0:
                cmapstr += pack(">H", cm) 
            else:
                try:
                    cmapstr += pack(">h", cm) 
                except:
                    warnings.warn("cmap value too big/small: %s" % cm)
                    cmapstr += pack(">H", -cm) 
        self.add('cmap', cmapstr)

        # glyf - Glyph data
        (glyfOffset,glyfLength) = self.get_table_pos('glyf')
        if (glyfLength < self.maxStrLenRead):
            glyphData = self.get_table('glyf')

        offsets = []
        glyf = b('')
        pos = 0

        hmtxstr = b('')
        xMinT = 0
        yMinT = 0
        xMaxT = 0
        yMaxT = 0
        advanceWidthMax = 0
        minLeftSideBearing = 0
        minRightSideBearing = 0
        xMaxExtent = 0
        maxPoints = 0            # points in non-compound glyph
        maxContours = 0            # contours in non-compound glyph
        maxComponentPoints = 0    # points in compound glyph
        maxComponentContours = 0    # contours in compound glyph
        maxComponentElements = 0    # number of glyphs referenced at top level
        maxComponentDepth = 0        # levels of recursion, set to 0 if font has only simple glyphs
        self.glyphdata = {}

        for originalGlyphIdx, uni in subsetglyphs: 
            # hmtx - Horizontal Metrics
            hm = self.getHMetric(orignHmetrics, originalGlyphIdx)    
            hmtxstr += hm

            offsets.append(pos)
            try:
                glyphPos = self.glyphPos[originalGlyphIdx]
                glyphLen = self.glyphPos[originalGlyphIdx + 1] - glyphPos
            except IndexError:
                warnings.warn("missing glyph %s" % (originalGlyphIdx))
                glyphLen = 0

            if (glyfLength < self.maxStrLenRead):
                data = substr(glyphData,glyphPos,glyphLen)
            else:
                if (glyphLen > 0):
                    data = self.get_chunk(glyfOffset+glyphPos,glyphLen)
                else:
                    data = b('')
            
            if (glyphLen > 0):
                up = unpack(">H", substr(data,0,2))[0]
            if (glyphLen > 2 and (up & (1 << 15)) ):     # If number of contours <= -1 i.e. composiste glyph
                pos_in_glyph = 10
                flags = GF_MORE
                nComponentElements = 0
                while (flags & GF_MORE):
                    nComponentElements += 1    # number of glyphs referenced at top level
                    up = unpack(">H", substr(data,pos_in_glyph,2))
                    flags = up[0]
                    up = unpack(">H", substr(data,pos_in_glyph+2,2))
                    glyphIdx = up[0]
                    self.glyphdata.setdefault(originalGlyphIdx, {}).setdefault('compGlyphs', []).append(glyphIdx)
                    try:
                        data = self._set_ushort(data, pos_in_glyph + 2, glyphSet[glyphIdx])
                    except KeyError:
                        data = 0
                        warnings.warn("missing glyph data %s" % glyphIdx)
                    pos_in_glyph += 4
                    if (flags & GF_WORDS): 
                        pos_in_glyph += 4 
                    else: 
                        pos_in_glyph += 2 
                    if (flags & GF_SCALE):
                        pos_in_glyph += 2 
                    elif (flags & GF_XYSCALE):
                        pos_in_glyph += 4 
                    elif (flags & GF_TWOBYTWO):
                        pos_in_glyph += 8 
                
                maxComponentElements = max(maxComponentElements, nComponentElements)
            
            glyf += data
            pos += glyphLen
            if (pos % 4 != 0): 
                padding = 4 - (pos % 4)
                glyf += str_repeat(b("\0"),padding)
                pos += padding

        offsets.append(pos)
        self.add('glyf', glyf)

        # hmtx - Horizontal Metrics
        self.add('hmtx', hmtxstr)

        # loca - Index to location
        locastr = b('')
        if (((pos + 1) >> 1) > 0xFFFF): 
            indexToLocFormat = 1        # long format
            for offset in offsets:
                locastr += pack(">L",offset) 
        else:
            indexToLocFormat = 0        # short format
            for offset in offsets:  
                locastr += pack(">H",int(offset/2)) 
        
        self.add('loca', locastr)

        # head - Font header
        head = self.get_table('head')
        head = self._set_ushort(head, 50, indexToLocFormat)
        self.add('head', head)

        # hhea - Horizontal Header
        hhea = self.get_table('hhea')
        hhea = self._set_ushort(hhea, 34, numberOfHMetrics)
        self.add('hhea', hhea)

        # maxp - Maximum Profile
        maxp = self.get_table('maxp')
        maxp = self._set_ushort(maxp, 4, numGlyphs)
        self.add('maxp', maxp)

        # OS/2 - OS/2
        os2 = self.get_table('OS/2')
        self.add('OS/2', os2 )

        self.fh.close()

        # Put the TTF file together
        stm = self.endTTFile('')
        return stm 
    

    #########################################
    # Recursively get composite glyph data
    def getGlyphData(self, originalGlyphIdx, nonlocals):
        # &maxdepth, &depth, &points, &contours
        nonlocals['depth'] += 1
        nonlocals['maxdepth'] = max(nonlocals['maxdepth'], nonlocals['depth'])
        if (len(self.glyphdata[originalGlyphIdx]['compGlyphs'])):
            for glyphIdx in self.glyphdata[originalGlyphIdx]['compGlyphs']: 
                self.getGlyphData(glyphIdx, nonlocals)            
        
        elif ((self.glyphdata[originalGlyphIdx]['nContours'] > 0) and nonlocals['depth'] > 0):     # simple
            contours += self.glyphdata[originalGlyphIdx]['nContours']
            points += self.glyphdata[originalGlyphIdx]['nPoints']
        
        nonlocals['depth'] -= 1


    #########################################
    # Recursively get composite glyphs
    def getGlyphs(self, originalGlyphIdx, nonlocals):
        # &start, &glyphSet, &subsetglyphs) 
        
        try:
            glyphPos = self.glyphPos[originalGlyphIdx]
            glyphLen = self.glyphPos[originalGlyphIdx + 1] - glyphPos
        except IndexError:
            warnings.warn("missing glyph %s" % (originalGlyphIdx))
            return

        if (not glyphLen):  
            return
        
        self.seek(nonlocals['start'] + glyphPos)
        numberOfContours = self.read_short()
        if (numberOfContours < 0):
            self.skip(8)
            flags = GF_MORE
            while (flags & GF_MORE): 
                flags = self.read_ushort()
                glyphIdx = self.read_ushort()
                if (glyphIdx not in nonlocals['glyphSet']):
                    nonlocals['glyphSet'][glyphIdx] = len(nonlocals['subsetglyphs'])    # old glyphID to new glyphID
                    nonlocals['subsetglyphs'].append((glyphIdx, 1))
                
                savepos = self.fh.tell()
                self.getGlyphs(glyphIdx, nonlocals)
                self.seek(savepos)
                if (flags & GF_WORDS):
                    self.skip(4)
                else:
                    self.skip(2)
                if (flags & GF_SCALE):
                    self.skip(2)
                elif (flags & GF_XYSCALE):
                    self.skip(4)
                elif (flags & GF_TWOBYTWO):
                    self.skip(8)

    #########################################

    def getHMTX(self, numberOfHMetrics, numGlyphs, glyphToChar, scale):
        start = self.seek_table("hmtx")
        aw = 0
        self.charWidths = [0] * 256*256
        nCharWidths = 0
        if ((numberOfHMetrics*4) < self.maxStrLenRead): 
            data = self.get_chunk(start,(numberOfHMetrics*4))
            arr = unpack(">%dH" % (int(len(data)/2)), data)
        else:
            self.seek(start) 
        for glyph in range(numberOfHMetrics): 
            if ((numberOfHMetrics*4) < self.maxStrLenRead):
                aw = arr[(glyph*2)] # PHP starts arrays from index 0!? +1
            else:
                aw = self.read_ushort()
                lsb = self.read_ushort()
            
            if (glyph in glyphToChar or glyph == 0):
                if (aw >= (1 << 15) ):
                    aw = 0     # 1.03 Some (arabic) fonts have -ve values for width
                    # although should be unsigned value - comes out as e.g. 65108 (intended -50)
                if (glyph == 0): 
                    self.defaultWidth = scale*aw
                    continue
                
                for char in glyphToChar[glyph]: 
                    if (char != 0 and char != 65535): 
                        w = int(round(scale*aw+0.001))   # ROUND_HALF_UP in PY3K (like php)
                        if (w == 0):  w = 65535 
                        if (char < 196608): 
                            self.charWidths[char] = w 
                            nCharWidths += 1
            
        
        data = self.get_chunk((start+numberOfHMetrics*4),(numGlyphs*2))
        arr = unpack(">%dH" % (int(len(data)/2)), data)
        diff = numGlyphs-numberOfHMetrics
        for pos in range(diff): 
            glyph = pos + numberOfHMetrics
            if (glyph in glyphToChar): 
                for char in glyphToChar[glyph]: 
                    if (char != 0 and char != 65535): 
                        w = int(round(scale*aw+0.001))  # ROUND_HALF_UP in PY3K (like php)
                        if (w == 0):  w = 65535 
                        if (char < 196608):
                            self.charWidths[char] = w
                            nCharWidths += 1 
                        
        
        # NB 65535 is a set width of 0
        # First bytes define number of chars in font
        self.charWidths[0] = nCharWidths 
    

    def getHMetric(self, numberOfHMetrics, gid): 
        start = self.seek_table("hmtx")
        if (gid < numberOfHMetrics):
            self.seek(start+(gid*4))
            hm = self.fh.read(4)
        else:
            self.seek(start+((numberOfHMetrics-1)*4))
            hm = self.fh.read(2)
            self.seek(start+(numberOfHMetrics*2)+(gid*2))
            hm += self.fh.read(2)
        return hm
    

    def getLOCA(self, indexToLocFormat, numGlyphs): 
        start = self.seek_table('loca')
        self.glyphPos = []
        if (indexToLocFormat == 0):
            data = self.get_chunk(start,(numGlyphs*2)+2)
            arr = unpack(">%dH" % (int(len(data)/2)), data)
            for n in range(numGlyphs): 
                self.glyphPos.append((arr[n] * 2))  # n+1 !?
        elif (indexToLocFormat == 1):
            data = self.get_chunk(start,(numGlyphs*4)+4)
            arr = unpack(">%dL" % (int(len(data)/4)), data)
            for n in range(numGlyphs):
                self.glyphPos.append((arr[n]))  # n+1 !?
        else:
            die('Unknown location table format ' + indexToLocFormat)

    # CMAP Format 4
    def getCMAP4(self, unicode_cmap_offset, glyphToChar, charToGlyph):
        self.maxUniChar = 0
        self.seek(unicode_cmap_offset + 2)
        length = self.read_ushort()
        limit = unicode_cmap_offset + length
        self.skip(2)

        segCount = int(self.read_ushort() / 2)
        self.skip(6)
        endCount = []
        for i in range(segCount):
            endCount.append(self.read_ushort())
        self.skip(2)
        startCount = []
        for i in range(segCount):
            startCount.append(self.read_ushort()) 
        idDelta = []
        for i in range(segCount):
            idDelta.append(self.read_short())         # ???? was unsigned short
        idRangeOffset_start = self._pos
        idRangeOffset = []
        for i in range(segCount):
            idRangeOffset.append(self.read_ushort()) 

        for n in range(segCount): 
            endpoint = (endCount[n] + 1)
            for unichar in range(startCount[n], endpoint, 1): 
                if (idRangeOffset[n] == 0):
                    glyph = (unichar + idDelta[n]) & 0xFFFF
                else:
                    offset = (unichar - startCount[n]) * 2 + idRangeOffset[n]
                    offset = idRangeOffset_start + 2 * n + offset
                    if (offset >= limit):
                        glyph = 0
                    else:
                        glyph = self.get_ushort(offset)
                        if (glyph != 0):
                           glyph = (glyph + idDelta[n]) & 0xFFFF
                    
                charToGlyph[unichar] = glyph
                if (unichar < 196608):
                    self.maxUniChar = max(unichar,self.maxUniChar) 
                glyphToChar.setdefault(glyph, []).append(unichar)

    # CMAP Format 12
    def getCMAP12(self, unicode_cmap_offset, glyphToChar, charToGlyph):
        self.maxUniChar = 0
        # table (skip format version, should be 12)
        self.seek(unicode_cmap_offset + 2)
        # reserved
        self.skip(2)
        # table length
        length = self.read_ulong()
        # language (should be 0)
        self.skip(4)
        # groups count
        grpCount = self.read_ulong()

        if 2 + 2 + 4 + 4 + 4 + grpCount * 3 * 4 > length:
            die("TTF format 12 cmap table too small")  
        for n in range(grpCount):
            startCharCode = self.read_ulong()
            endCharCode = self.read_ulong()
            glyph = self.read_ulong()
            for unichar in range(startCharCode, endCharCode + 1):
                charToGlyph[unichar] = glyph
                if (unichar < 196608):
                    self.maxUniChar = max(unichar, self.maxUniChar) 
                glyphToChar.setdefault(glyph, []).append(unichar)
                glyph += 1
            
            

    # Put the TTF file together
    def endTTFile(self, stm): 
        stm = b('')
        numTables = count(self.otables)
        searchRange = 1
        entrySelector = 0
        while (searchRange * 2 <= numTables): 
            searchRange = searchRange * 2
            entrySelector = entrySelector + 1
        
        searchRange = searchRange * 16
        rangeShift = numTables * 16 - searchRange

        # Header
        if (_TTF_MAC_HEADER): 
            stm += (pack(">LHHHH", 0x74727565, numTables, searchRange, entrySelector, rangeShift))    # Mac
        else:
            stm += (pack(">LHHHH", 0x00010000 , numTables, searchRange, entrySelector, rangeShift))    # Windows

        
        # Table directory
        tables = self.otables

        offset = 12 + numTables * 16
        sorted_tables = sorted(tables.items())
        for tag, data in sorted_tables:
            if (tag == 'head'):
                head_start = offset 
            stm += tag.encode("latin1")
            checksum = calcChecksum(data)
            stm += pack(">HH", checksum[0],checksum[1])
            stm += pack(">LL", offset, strlen(data))
            paddedLength = (strlen(data)+3)&~3
            offset = offset + paddedLength

        # Table data
        for tag, data in sorted_tables: 
            data += b("\0\0\0")
            stm += substr(data,0,(strlen(data)&~3))

        checksum = calcChecksum(stm)
        checksum = sub32((0xB1B0,0xAFBA), checksum)
        chk = pack(">HH", checksum[0],checksum[1])
        stm = self.splice(stm,(head_start + 8),chk)
        return stm 
    



# *-* coding: utf-8 *-*
import sys
import os
from email.mime.application import MIMEApplication

from asn1crypto import cms, core, algos
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.asymmetric import padding

from endesive import signer


class EncryptedData(object):

    def email(self, data, oaep):
        prefix = ['x-', ''][oaep]
        msg = MIMEApplication(data)
        del msg['Content-Type']
        msg['Content-Disposition'] = 'attachment; filename="smime.p7m"'
        msg['Content-Type'] = 'application/%spkcs7-mime; smime-type=enveloped-data; name="smime.p7m"' % prefix

        data = msg.as_string()
        return data

    def pad(self, s, block_size):
        n = block_size - len(s) % block_size
        if n == 0:
            n = block_size
            #return s
        n = bytes([n] * n)
        return s + n

    def recipient_info(self, cert, session_key, oaep):
        public_key = cert.public_key()
        cert = signer.cert2asn(cert)

        tbs_cert = cert['tbs_certificate']
        # TODO: use subject_key_identifier when available
        if oaep:
            encrypted_key = public_key.encrypt(
                session_key,
                padding.OAEP(
                    mgf=padding.MGF1(hashes.SHA512()),
                    algorithm=hashes.SHA512(),
                    label=None
                )
            )
            kea = cms.KeyEncryptionAlgorithm({
                'algorithm': cms.KeyEncryptionAlgorithmId('rsaes_oaep'),
                'parameters': algos.RSAESOAEPParams({
                    'hash_algorithm': algos.DigestAlgorithm({'algorithm': 'sha512'}),
                    'mask_gen_algorithm': algos.MaskGenAlgorithm({
                        'algorithm': algos.MaskGenAlgorithmId('mgf1'),
                        'parameters': {
                            'algorithm': algos.DigestAlgorithmId('sha512'),
                        }
                    }),
                    'p_source_algorithm': algos.PSourceAlgorithm({
                        'algorithm': algos.PSourceAlgorithmId('p_specified'),
                        'parameters': b'',
                    })
                })
            })
        else:
            kea = {'algorithm': 'rsa'}
            encrypted_key = public_key.encrypt(session_key, padding.PKCS1v15())
        result = cms.RecipientInfo(
            name='ktri',
            value={
                'version': 'v0',
                'rid': cms.RecipientIdentifier(
                    name='issuer_and_serial_number',
                    value={
                        'issuer': tbs_cert['issuer'],
                        'serial_number': tbs_cert['serial_number']
                    }
                ),
                'key_encryption_algorithm': kea,
                'encrypted_key': core.OctetString(encrypted_key)
            }
        )
        return result

    def build(self, data, certs, algo, oaep):
        key_size = {
            'aes128': 16,
            'aes192': 24,
            'aes256': 32,
        }[algo.split('_', 1)[0]]
        block_size = 16
        session_key = os.urandom(key_size)
        iv = os.urandom(block_size)
        cipher = Cipher(
            algorithms.AES(session_key),
            getattr(modes, algo.split('_', 1)[1].upper())(iv),
            default_backend()
        )

        data = self.pad(data, block_size)

        encryptor = cipher.encryptor()
        data = encryptor.update(data) + encryptor.finalize()

        recipient_infos = []
        for cert in certs:
            recipient_info = self.recipient_info(cert, session_key, oaep)
            recipient_infos.append(recipient_info)

        enveloped_data = cms.ContentInfo({
            'content_type': 'enveloped_data',
            'content': {
                'version': 'v0',
                'recipient_infos': recipient_infos,
                'encrypted_content_info': {
                    'content_type': 'data',
                    'content_encryption_algorithm': {
                        'algorithm': algo,
                        'parameters': iv
                    },
                    'encrypted_content': data
                }
            }
        })
        data = self.email(enveloped_data.dump(), oaep)
        return data


def encrypt(data, certs, algo='aes256_cbc', oaep=False):
    assert algo[:3] == 'aes' and algo.split('_', 1)[1] in ('cbc', 'ofb')
    cls = EncryptedData()
    return cls.build(data, certs, algo, oaep)
